import math
import numpy as np
from tqdm import tqdm
import os, sys
import itertools
import time
import torchvision
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
from torchvision import transforms
from collections import defaultdict
import sh

sys.path.append('../')

from src.optimization import SFW, PositiveKSparsePolytope
from torch.distributions import Categorical

from params_file import *

from src.basic_nn_classes import Classifier, SimpleNet

from src.models import get_model
from src.utilities import AddMaskChannel, Averager, categorical_accuracy

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import argparse


from src.arthur_merlin_networks import opt_mask_NN_FW, train_mask_epoch, evaluate_mask_model
from src.arthur_merlin_optimizers import opt_mask
from src.arthur_merlin_utilities import set_labels, Metrics_collector
from src.basic_training import regular_training, evaluate_model, evaluate_reg_model, train_reg_epoch
from src.arthur_merlin_plotting import show_MnM_images, show_example_images


parser = argparse.ArgumentParser()

parser.add_argument('--epocs', help='Number of epochs to run', default=default_epocs, type=int)
parser.add_argument('--bs', help='Batch Size', default=default_bs, type=int)

parser.add_argument('--lr', help='Learning Rate', default=default_lr, type=float)
parser.add_argument('--en', help='Expected Noise', default=default_en, type=float)
parser.add_argument('--mt', help='Model Type', default=default_mt, type=str)

parser.add_argument('--gamma', help='Parameter to tune the objective function', default=default_gamma, type=float)

parser.add_argument('--merlin_nn', help='is merlin a NN', type=int)
parser.add_argument('--morgana_nn', help='is morgana a NN', type=int)

args = parser.parse_args()

main_direc = f"./"


torch.cuda.is_available()

batch_size = args.bs
learning_rate = args.lr
expected_noise = args.en
model_type = args.mt
nepochs = args.epocs
gamma = args.gamma

#These two parameters control whether merlin and morgana use a NN or the optimizer

if args.merlin_nn == 1:
    use_NN_merlin = True
else:
    use_NN_merlin = False

if args.morgana_nn == 1:
    use_NN_morgana = True
else:
    use_NN_morgana = False

nclasses = len(classes)+1

print("completed step 1")

params={}
params['expected_noise'] = expected_noise
params['nclasses'] = nclasses
params['device'] = device


transformChain = transforms.Compose([transforms.ToTensor(), AddMaskChannel()])
trainData = torchvision.datasets.MNIST(root='.', train=True, download=True, transform=transformChain)
testData = torchvision.datasets.MNIST(root='.', train=False, download=True, transform=transformChain)

if restrict_size == True:
    trainData.data = trainData.data[:1000]
    trainData.targets = trainData.targets[:1000]


trainData, testData = set_labels(trainData, testData, classes)
    
trainLoader = torch.utils.data.DataLoader(trainData, batch_size=batch_size, shuffle=True, pin_memory=torch.cuda.is_available(), num_workers=2)
testLoader = torch.utils.data.DataLoader(testData, batch_size=batch_size, shuffle=False, pin_memory=torch.cuda.is_available(), num_workers=2)

loss_criterion = torch.nn.CrossEntropyLoss().to(device=device)
softmax = torch.nn.Softmax(dim=1).to(device=device)


model = get_model(model_type=model_type, nclasses=nclasses).to(device)
basemodel_path = main_direc + f"trained_models_merlin_morgana/p3-{model_type}-c{classes_str}.pt"

if (not os.path.isfile(basemodel_path)) or overwrite:
    optimizer = torch.optim.Adam(model.parameters(), lr=0.0001)
    regular_training(model, optimizer, 5, trainLoader, testLoader, device, loss_criterion, nclasses)
    torch.save(model.state_dict(), basemodel_path)
else:
    model.load_state_dict(torch.load(basemodel_path))
    

    
print("completed step 2")


def p3_training(masksizeList, model_type, nepochs, gamma=0.5, overwrite = False, max_iter=1500,
                use_NN_merlin=True, use_NN_morgana=True, pic_pdf = None, seed = None):
    
    """
    This function trains the merlin, morgana and arthur neural networks
    """
    # Initialize the arthur model
    model = get_model(model_type=model_type, nclasses=nclasses).to(device)
    #load the previously trained model
    basemodel_path = main_direc + f"trained_models_merlin_morgana/p3-{model_type}-c{classes_str}.pt"
    model.load_state_dict(torch.load(basemodel_path))
    
    #Initialize the arthur optimizer
    optimizer = torch.optim.Adam(model.parameters(), lr=0.0001)
    
    #Initialize the merlin and morgana models and optimizers depending on if they are going to be used
    if use_NN_merlin == True:
        merlin_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
        merlin_model_opt = torch.optim.Adam(merlin_model.parameters(), lr=1e-4)
    else:
        merlin_model = None
        merlin_model_opt = None

    if use_NN_morgana == True:
        morgana_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
        morgana_model_opt = torch.optim.Adam(morgana_model.parameters(), lr=1e-6)
    else:
        morgana_model = None
        morgana_model_opt = None
    
    #use_NN = True
    
    for mask_size in masksizeList:
        #These are the files names where the various model parameters are going to be stored
        trainedmodel_path = main_direc + f"trained_models_merlin_morgana/g{classes_str}_{model_type}_{mask_size}_g{gamma}_trained_{int(use_NN_merlin)}{int(use_NN_morgana)}_s{seed}.pt"
        trained_merlin_model_path = main_direc + f"trained_models_merlin_morgana/g{classes_str}_merlin_{model_type}_{mask_size}_g{gamma}_trained_{int(use_NN_merlin)}{int(use_NN_morgana)}_s{seed}.pt"
        trained_morgana_model_path = main_direc + f"trained_models_merlin_morgana/g{classes_str}_morgana_{model_type}_{mask_size}_g{gamma}_trained_{int(use_NN_merlin)}{int(use_NN_morgana)}_s{seed}.pt"
        
        #For each new mask size reload the initial arthur mask
        if (not os.path.isfile(trainedmodel_path)) or overwrite:
            model.load_state_dict(torch.load(basemodel_path))
            
            for epoch in range(0, nepochs+1):
                print(f"epoch:{epoch}")
                if epoch % 2 == 0:
                    # Train with Merlin and Morgana
                    print("Train with Merlin and Morgana")
                    train_collector = train_mask_epoch(model, trainLoader, use_NN_merlin, use_NN_morgana, merlin_model,
                                                       merlin_model_opt, morgana_model, morgana_model_opt, optimizer, 
                                                       k = mask_size, gamma=gamma, max_iter=max_iter, params=params, 
                                                       pic_pdf = pic_pdf)

                    #Here we evaluate the trained merlin and morgana masks
                    test_collector = evaluate_mask_model(model, use_NN_merlin, use_NN_morgana, merlin_model,
                                                         merlin_model_opt, morgana_model, morgana_model_opt,
                                                         loader=testLoader, gamma=gamma, mask_size=mask_size, max_iter=0,
                                                         params=params)

                    print("Mask Size:", mask_size, "Epoch:", epoch)
                    train_collector.show_results("TRAIN   ")
                    test_collector.show_results("TEST   ")
                    print("")

                    textfig = plt.figure(figsize=(22,8))

                    txt1 = f"MaskSize:,{mask_size}, Epoch:, {epoch}"
                    txt2 = train_collector.return_results("TRAIN   ")
                    txt3 = test_collector.return_results("TEST   ")

                    txt = txt1 + "\n" + txt2 + "\n" + txt3

                    textfig.text(0.5,0.5,txt, transform=textfig.transFigure, size=24, ha="center")
                    pic_pdf.savefig(textfig)
                    plt.close()

                    del train_collector, test_collector
                        
                #Train t
                if epoch % 2 == 1:
                    # Train normally i.e. without merlin and morgana
                    print("Train normal")
                    train_loss, train_acc = train_reg_epoch(model, optimizer, loader=trainLoader, params = params,
                                                           loss_criterion=loss_criterion)
                    test_loss, test_acc = evaluate_reg_model(model, loader=testLoader, params = params,
                                                           loss_criterion=loss_criterion)

                    print(f"TRAIN   loss: {train_loss:1.4f}, acc: {100*train_acc:1.2f}%")
                    print(f"TEST    loss: {test_loss:1.4f}, acc: {100*test_acc:1.2f}%")

                    txt1 = f"Mask Size:,{mask_size}, Epoch:, {epoch}"
                    txt2 = f"TRAIN   loss: {train_loss:1.4f}, acc: {100*train_acc:1.2f}%"
                    txt3 = f"TEST    loss: {test_loss:1.4f}, acc: {100*test_acc:1.2f}%"
                    print(txt1)
                    print(txt2)
                    print(txt3)
                    print("")

                    textfig = plt.figure(figsize=(22,8))

                    txt = txt1 + "\n" + txt2 + "\n" + txt3

                    textfig.text(0.5,0.5,txt, transform=textfig.transFigure, size=24, ha="center")
                    pic_pdf.savefig(textfig)
                    plt.close()

                    del train_loss, train_acc, test_loss, test_acc

            # Save the model
            torch.save(model.state_dict(), trainedmodel_path)
            
            #We save the networks only if they have been used
            if use_NN_merlin == True:
                torch.save(merlin_model.state_dict(), trained_merlin_model_path)
            
            if use_NN_morgana == True:
                torch.save(morgana_model.state_dict(), trained_morgana_model_path)
    
    return True

#Here we iterate over different values of gamma and train the morgana and merlin masks for each value of gamma

for seed in seeds:
    torch.manual_seed(seed)

    print("starting funcs")

    #This file creates the images and saves them to the file
    pic_pdf = PdfPages(main_direc + f"results_merlin_morgana/g{classes_str}-{model_type}-c{classes_str}-g{gamma}-trained_00_s{seed}.pdf")

    #This funciton trains the merlin, morgana and arthur
    p3_training(masksizeList, model_type = model_type, nepochs = nepochs, gamma=gamma, overwrite = True,
                max_iter=max_iter, use_NN_merlin=use_NN_merlin, use_NN_morgana=use_NN_morgana, pic_pdf = pic_pdf,
                seed = seed)

    pic_pdf.close()





