from src.arthur_merlin_utilities import Metrics_collector, merlin_criterion, morgana_criterion, get_distortion, check_stopping, get_accuracy, apply_mask
from src.optimization import SFW, PositiveKSparsePolytope
from src.arthur_merlin_plotting import show_MnM_images_MC, show_MnM_images
from src.arthur_merlin_optimizers import opt_mask

import torch.nn as nn
import torch
from tqdm import tqdm



def normalize_l1(s, k):
    factor = torch.clamp(k/(1e-7 + torch.norm(s, p=1, dim=(1,2), keepdim=True)), max=1)
    #print(f"range:{torch.min(torch.norm(s, p=1, dim=(1,2), keepdim=True))},{torch.max(torch.norm(s, p=1, dim=(1,2), keepdim=True))}")
    return factor*s

def mask_entropy(s):
    #return -(torch.xlogy(s,s).sum()+torch.xlogy(1-s,1-s).sum())
    return (s*(1-s)).sum()

def get_FW_direction(s, constraint):
    v = constraint.lmo(s.grad)
    #print(f"vertex_sum:{torch.min(torch.sum(v,dim=(1,2)))}")
    return constraint.lmo(s.grad) - s

def opt_mask_NN_FW(model, x_input, y_target, mode, k, mask, max_iter = 5, 
                   lr=0.1, momentum=0.9, stoptol = 1e-5, verbose=False, penalty=0.25,
                   mask_model=None, mask_model_opt=None, test_mode = False, params=None):
    """
    This function is supposed to replace the basic mask models with neural networks
    here the either mask or both can be generated by either the FW approach or the neural netowrk model
    arguments:
    use_NNs: boolean tuple which indicates whether to use the NN model for merlin and morgana respectively
    """
    
    device = params['device']
    # Put input and target on GPU
    x_input = x_input.to(device)
    y_target = y_target.to(device)
    batch_size = y_target.shape[0]
    assert x_input.numel() % batch_size == 0

    # Set constraint for optimisation
    constraint = PositiveKSparsePolytope(n=x_input.numel()/batch_size, bs=batch_size, k=k)
    
    # Do not push nn.Parameter() to cuda, since this makes s non-leaf
    s = nn.Parameter(mask.to(device), requires_grad=True)
    s = constraint.shift_inside(s, check_necessity=False) # NOT ACTIVATED: Will only do something if necessary
    s.requires_grad = True
        
    if mode=='merlin':
        criterion = merlin_criterion
    elif mode=='morgana':
        criterion = morgana_criterion
    
    #rde_optim = SFW([s], learning_rate=lr, momentum=momentum, rescale=None)
    
    # Disable the network's autograd
    for param in model.parameters():
         param.requires_grad = False
    
    # Early Stopping variables:
    dist_old = float('inf')
    stop = False
    it = 0 if verbose else 1
    metrics = dict()
    
    exp_noise = params['expected_noise']
    # Optimisation Iteration:
    
    while not stop:
        if test_mode == True:
            #print("testing")
            s = nn.Parameter(normalize_l1(mask_model.eval()(x_input.unsqueeze(1)),k).detach(), requires_grad=True)
            break
        
        # Zero grads
        mask_model_opt.zero_grad()
        s = nn.Parameter(normalize_l1(mask_model.train()(x_input.unsqueeze(1)),k).detach(), requires_grad=True)

        distortion = get_distortion(model=model, x=x_input, y=y_target, s=s, criterion=criterion, en=exp_noise)
        regularizer = penalty*torch.mean(torch.abs(s))
        loss = distortion + regularizer

        loss.backward()

        with torch.no_grad():
            #FW_update = get_FW_direction(s, constraint)
            FW_update = -s.grad

        mask_model_opt.zero_grad()

        #Same for both merlin and morgana
        mask_loss = -torch.sum(FW_update*normalize_l1(mask_model.train()(x_input.unsqueeze(1)),k))

        mask_loss.backward()    
   
        
        # Training Step
        if it != 0:
            #print("taking grad step")
            mask_model_opt.step()
                 
        # Give Messages
        if verbose and (it % 100 == 0 or it == 1):
            opt_feedback(model, x_input, y_target, mode, s, constraint, it, loss.item(), distortion.item())
        
        
        # Check stopping criteria:
        if it != 0:
            stop = check_stopping(it, max_iter, distortion.item(), dist_old, verbose, stoptol)
            dist_old = distortion.item()
            
        #Count Steps
        it += 1
    
    with torch.no_grad():
        s = constraint.shift_inside(s, check_necessity=False) # Will get the closest vertices
        
        # Set metrics
        distortion = get_distortion(model=model, x=x_input, y=y_target, s=s, criterion=criterion, en=exp_noise).item()
        delta_distortion = abs(dist_old - distortion)
        acc = get_accuracy(model, mode, x_input, y_target, s, exp_noise)
        l_0 = (s > 1e-30).sum().item()
        sparsity = 1 - l_0/float(s.numel())
        
        metrics['it_until_conv'] = it - 1
        metrics['delta_distortion'] = delta_distortion
        metrics['distortion'] = distortion
        metrics['sparsity'] = sparsity
        metrics['accuracy'] = acc    

    # Enable the network's autograd
    for param in model.parameters():
         param.requires_grad = True
            
    return (apply_mask(x_input, s, exp_noise).detach().clone()), s.detach().clone(), metrics
    # returns mask, masked images metrics
    

def train_mask_epoch(model, trainLoader, use_NN_merlin, use_NN_morgana, merlin_model, merlin_model_opt, 
                     morgana_model, morgana_model_opt, optimizer, gamma=False, k = False, max_iter=200, 
                     params=None, pic_pdf=None):
    """
    This function trains the merlin, morgan and arthur networks or optimizers for 1 epoch
    
    model: NN model for arthur
    trainLoader: training data
    use_NN_merlin: do we use a NN model for merlin
    use_NN_morgana: do we use a NN model for morgana
    merlin_model: NN model for merlin if it is to be used
    merlin_model_opt: Optimizer for NN model for merlin
    morgana_model: NN model for morgana if it is to be used
    morgana_model_opt: Optimizer for NN model for morgana
    optimizer: optimizer for arthur
    gamma: parameter which controls the weight assigned to the morgana loss function
    max_iter: max iterations for the FW solver or the number of times we go over the same data point for a NN
    params: dictionary of various parameters used in various locations
    pic_pdf: file to store images 
    """
    
    assert k and gamma
    
    metrics_collector = Metrics_collector()
    #count = 0
    
    device = params['device']
    
    for idx, (x_input, y_target) in enumerate(trainLoader):
        if idx%10 == 0:
            print(f"Currently at batch:{idx}")
            
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        #Extract initial mask
        mask = x_input.select(dim=1, index=1)
        
        
        #Obtain initial masks (merlin and morgana) by evaluating the current NN or by using the solver
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=k, mask=mask, 
                                                                stoptol=1e-5, max_iter=0, mask_model = merlin_model,
                                                                mask_model_opt = merlin_model_opt, params=params, 
                                                                test_mode = True)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=k, mask=mask, max_iter=200,
                                                          params=params)
            
        
        
        if use_NN_morgana == True:
            x_morgana, s_morgana, morgana_metrics = opt_mask_NN_FW(model, x_image, y_target, 'morgana', k=k, mask=mask, 
                                                                   stoptol=1e-5, max_iter=0, mask_model = morgana_model, 
                                                                   mask_model_opt = morgana_model_opt, params=params, 
                                                                   test_mode=True)
        else:
            x_morgana, s_morgana, morgana_metrics = opt_mask(model, x_image, y_target, 'morgana', k=k, mask=mask, max_iter=200,
                                                             params=params)
        
 
        
        optimizer.zero_grad()
        merlin_output = model.train()(x_merlin)
        morgana_output = model.train()(x_morgana)
        loss = merlin_criterion(merlin_output, y_target) - gamma*morgana_criterion(morgana_output, y_target)
       
        loss.backward()
        optimizer.step()
        
        
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=k, mask=mask, 
                                                                stoptol=1e-5, max_iter=max_iter, mask_model = merlin_model,
                                                                mask_model_opt = merlin_model_opt, params=params)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=k, mask=mask, max_iter=200,
                                                          params=params)
            
        
        
        if use_NN_morgana == True:
            x_morgana, s_morgana, morgana_metrics = opt_mask_NN_FW(model, x_image, y_target, 'morgana', k=k, mask=mask, 
                                                                   stoptol=1e-5, max_iter=max_iter, mask_model = morgana_model, 
                                                                   mask_model_opt = morgana_model_opt, params=params)
        else:
            x_morgana, s_morgana, morgana_metrics = opt_mask(model, x_image, y_target, 'morgana', k=k, mask=mask, max_iter=200,
                                                             params=params)


        assert x_merlin.grad == None
        assert s_merlin.grad == None
        assert x_morgana.grad == None
        assert s_morgana.grad == None
        #print(count)
        #count+=1

        metrics_collector.add(loss.item(), merlin_metrics, morgana_metrics)

    show_MnM_images(x_image, x_merlin, s_merlin, x_morgana, s_morgana, n_samples = 8, pic_pdf = pic_pdf)    
    
    return metrics_collector


def evaluate_mask_model(model, use_NN_merlin, use_NN_morgana, merlin_model, merlin_model_opt, morgana_model,
                        morgana_model_opt, loader, gamma, mask_size = None, max_iter=1500, params=None):
    assert mask_size
    
    device = params['device']
    metrics_collector = Metrics_collector()
    #count = 0
    for x_input, y_target in loader:
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        mask = x_input.select(dim=1, index=1)
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                                stoptol=1e-5, max_iter=max_iter, mask_model = merlin_model, 
                                                                mask_model_opt = merlin_model_opt, test_mode = True,
                                                                params=params)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                          max_iter=200, params=params)

        if use_NN_morgana == True:
            x_morgana, s_morgana, morgana_metrics = opt_mask_NN_FW(model, x_image, y_target, 'morgana', k=mask_size, mask=mask,
                                                                   stoptol=1e-5, max_iter=4*max_iter, 
                                                                   mask_model = morgana_model,
                                                                   mask_model_opt = morgana_model_opt, test_mode = True,
                                                                   params=params)
        else:
            x_morgana, s_morgana, morgana_metrics = opt_mask(model, x_image, y_target, 'morgana', k=mask_size, mask=mask,
                                                             max_iter=200, params=params)
                
        
        with torch.no_grad():
            merlin_output = model.train()(x_merlin)
            morgana_output = model.train()(x_morgana)
            loss = merlin_criterion(merlin_output, y_target) - gamma*morgana_criterion(morgana_output, y_target)

        metrics_collector.add(loss.item(), merlin_metrics, morgana_metrics)
    return metrics_collector


def train_merlin_only_mask_epoch(model, trainLoader, use_NN_merlin, merlin_model, merlin_model_opt,
                                 optimizer, k = False, max_iter=1500, params=None, pic_pdf=None):
    """
    Train a merlin only mask for 1 epoch
    """
    assert k
    
    metrics_collector = Metrics_collector()
    #count = 0
    
    device = params['device']
    
    for idx, (x_input, y_target) in enumerate(trainLoader):
        if idx%10 == 0:
            print(f"Currently at batch:{idx}")
            
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        mask = x_input.select(dim=1, index=1)
        
        
        
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=k, mask=mask, 
                                                                stoptol=1e-5, max_iter=0, mask_model = merlin_model,
                                                                mask_model_opt = merlin_model_opt, params=params, 
                                                                test_mode = True)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=k, mask=mask, max_iter=500,
                                                          params=params)
            
        
 
        
        optimizer.zero_grad()
        merlin_output = model.train()(x_merlin)
        loss = merlin_criterion(merlin_output, y_target)
       
        loss.backward()
        optimizer.step()
        
        
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=k, mask=mask, 
                                                                stoptol=1e-5, max_iter=max_iter, mask_model = merlin_model,
                                                                mask_model_opt = merlin_model_opt, params=params)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=k, mask=mask, max_iter=500,
                                                          params=params)


        assert x_merlin.grad == None
        assert s_merlin.grad == None

        #print(count)
        #count+=1

        metrics_collector.add_merlin_only(loss.item(), merlin_metrics)

    #show_MnM_images(x_image, x_merlin, s_merlin, x_morgana, s_morgana, n_samples = 8, pic_pdf = pic_pdf)    
    
    return metrics_collector

def evaluate_merlin_only_mask_model(model, use_NN_merlin, merlin_model, merlin_model_opt, loader, mask_size = None, 
                                    max_iter=1500, params=None):
    assert mask_size
    
    device = params['device']
    metrics_collector = Metrics_collector()
    #count = 0
    for x_input, y_target in loader:
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        mask = x_input.select(dim=1, index=1)
        if use_NN_merlin == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                                stoptol=1e-5, max_iter=max_iter, mask_model = merlin_model, 
                                                                mask_model_opt = merlin_model_opt, test_mode = True,
                                                                params=params)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                          max_iter=500, params=params)

        with torch.no_grad():
            merlin_output = model.train()(x_merlin)
            loss = merlin_criterion(merlin_output, y_target) 

        metrics_collector.add_merlin_only(loss.item(), merlin_metrics)
    return metrics_collector
