from src.arthur_merlin_utilities import Metrics_collector, merlin_criterion, morgana_criterion, get_distortion, check_stopping, get_accuracy, apply_mask 
from src.optimization import SFW, PositiveKSparsePolytope
import torch.nn as nn
import torch

def opt_mask(model, x_input, y_target, mode, k, mask, max_iter = 1500, lr=0.1, momentum=0.9, stoptol = 1e-5, verbose=False, penalty=0.25, params=None):
    
    device = params['device']
    # Put input and target on GPU
    x_input = x_input.to(device)
    y_target = y_target.to(device)
    batch_size = y_target.shape[0]
    assert x_input.numel() % batch_size == 0

    # Set constraint for optimisation
    constraint = PositiveKSparsePolytope(n=x_input.numel()/batch_size, bs=batch_size, k=k)
    
    # Do not push nn.Parameter() to cuda, since this makes s non-leaf
    s = nn.Parameter(mask.to(device), requires_grad=True)
    s = constraint.shift_inside(s, check_necessity=False) # NOT ACTIVATED: Will only do something if necessary
    s.requires_grad = True
    
    if mode=='merlin':
        criterion = merlin_criterion
    elif mode=='morgana':
        criterion = morgana_criterion
    rde_optim = SFW([s], learning_rate=lr, momentum=momentum, rescale=None)
    
    # Disable the network's autograd
    for param in model.parameters():
         param.requires_grad = False
    
    # Early Stopping variables:
    dist_old = float('inf')
    stop = False
    it = 0 if verbose else 1
    metrics = dict()
    
    en = params['expected_noise']
    # Optimisation Iteration:
    
    while not stop:
        # Zero grads
        rde_optim.zero_grad()
        
        # Compute objective
        distortion = get_distortion(model=model, x=x_input, y=y_target, s=s, criterion=criterion, en=en)
        regularizer = penalty*torch.mean(torch.abs(s))
        loss = distortion + regularizer
        
        loss.backward()
        
        # Training Step
        if it != 0:
            rde_optim.step(constraints=[constraint])
        
        # Give Messages
        if verbose and (it % 100 == 0 or it == 1):
            opt_feedback(model, x_input, y_target, mode, s, constraint, it, loss.item(), distortion.item())
        
        
        # Check stopping criteria:
        if it != 0:
            stop = check_stopping(it, max_iter, distortion.item(), dist_old, verbose, stoptol)
            dist_old = distortion.item()
            
        #Count Steps
        it += 1
    
    with torch.no_grad():
        s = constraint.shift_inside(s, check_necessity=False) # Will get the closest vertices
        
        # Set metrics
        distortion = get_distortion(model=model, x=x_input, y=y_target, s=s, criterion=criterion, en=en).item()
        delta_distortion = abs(dist_old - distortion)
        acc = get_accuracy(model, mode, x_input, y_target, s, en=en)
        l_0 = (s > 1e-30).sum().item()
        sparsity = 1 - l_0/float(s.numel())
        
        metrics['it_until_conv'] = it - 1
        metrics['delta_distortion'] = delta_distortion
        metrics['distortion'] = distortion
        metrics['sparsity'] = sparsity
        metrics['accuracy'] = acc    

    # Enable the network's autograd
    for param in model.parameters():
         param.requires_grad = True
            
    return (apply_mask(x_input, s, en).detach().clone()), s.detach().clone(), metrics
    # returns mask, masked images metrics