import torch
from src.utilities import Averager
import copy

softmax = torch.nn.Softmax(dim=1)

def set_labels(trainData, testData, classes):
    label_count = 1

    trainData.targets = torch.tensor(trainData.targets)
    testData.targets = torch.tensor(testData.targets)

    train_idxs = torch.zeros_like(trainData.targets)
    test_idxs = torch.zeros_like(testData.targets)
    
    for label in classes:    
        label_index = label
        #label_index = trainData.classes.index(label)
        train_idxs = torch.logical_or(train_idxs, trainData.targets==label_index)
        test_idxs = torch.logical_or(test_idxs, testData.targets==label_index)
        trainData.targets[trainData.targets==label_index] = label_count
        testData.targets[testData.targets==label_index] = label_count
        label_count += 1
    trainData.data = trainData.data[train_idxs]
    trainData.targets = trainData.targets[train_idxs]
    testData.data = testData.data[test_idxs]
    testData.targets = testData.targets[test_idxs]
    
    return trainData, testData


def set_labels_data(testData, classes):
    td = copy.deepcopy(testData)
    label_count = 1
    test_idxs = torch.zeros_like(td.targets)
    
    for label in classes:    
        test_idxs = torch.logical_or(test_idxs, td.targets==label)
        #td.targets[testData.targets==label] = label
        label_count += 1
        
    td.data = td.data[test_idxs]
    td.targets = td.targets[test_idxs]
    
    return td

def get_distortion(model, x, y, s, criterion, en):
    x_masked = apply_mask(x,s,en)
    output = model.eval()(x_masked)
    return criterion(output, y)

def get_random_distortion(model, x, y, s, criterion, en):
    x_masked = apply_random_mask(x,s,en)
    output = model.eval()(x_masked)
    return criterion(output, y)

"""
Creating expansion matrices
This function creates new matrices by repeating 1s and 0s
These new matrices are then used in the construction of a larger mask
by duplicating elements of smaller mask
"""
def create_expan_matrix(mask_shape, image_shape):
    if (image_shape[0]%mask_shape[0] != 0) or (image_shape[1]%mask_shape[1] != 0):
        raise ValueError("mask and image shapes don't match")

    P1 = torch.zeros(mask_shape[1], image_shape[1])
    P2 = torch.zeros(image_shape[0], mask_shape[0])

    msr = int(image_shape[0]/mask_shape[0])
    msc = int(image_shape[1]/mask_shape[1])

    for j in range(mask_shape[1]):
        P1[j,msc*j:msc*j+msc] = 1.

    for j in range(mask_shape[0]):
        P2[msr*j:msr*j+msr,j] = 1.

    return P1, P2

"""
Use a mask with fixed noise
single_mask_size: mask size if there is a single mask
"""
def apply_mask(x, s, en):
    expected_noise = en
    if s.shape[2:] == x.shape[2:]:
        return s*x + expected_noise*(1-s)*torch.ones_like(x)
    else:
        p1, p2 = create_expan_matrix(s.shape[2:], x.shape[2:])
        p1 = p1.unsqueeze(0)
        p1 = p1.unsqueeze(1)
        p2 = p2.unsqueeze(0)
        p2 = p2.unsqueeze(1)
        p1 = p1.to(x.device)
        p2 = p2.to(x.device)

        exp_s = torch.matmul(p2, torch.matmul(s, p1)).to(x.device)
        return exp_s*x + expected_noise*(1-exp_s)*torch.ones_like(x) 


"""
Use a mask with random noise
"""
def apply_random_mask(x, s, en):
    expected_noise = torch.maximum(en + 0.05 * torch.randn(x.shape), torch.zeros(x.shape)).to(x.device)
    return s*x + expected_noise*(1-s)*torch.ones_like(x)

def get_accuracy(model, mode, x, y, s, en, constraint=False):
    if constraint:
        s = constraint.shift_inside(s, check_necessity=False)
    x_masked = apply_mask(x,s,en)
    prediction = torch.argmax(model.eval()(x_masked), dim=1)
    if mode=='merlin':
        accuracy=prediction.eq(y.squeeze()).sum().item()/float(len(y))
    elif mode=='morgana':
        accuracy = torch.logical_or(prediction.eq(y.squeeze()),prediction.eq(0)).sum().item()/float(len(y))
    return accuracy


def opt_feedback(model, x_input, y_target, mode, s, constraint, it, loss, distortion):
    with torch.no_grad():
        l_0 = (s > 1e-30).sum().item()
        sparsity = 1 - l_0/float(s.numel())
        l_1 = (torch.norm(s, p=1)/float(s.numel())).item()
        
        acc_unc = get_accuracy(model=model, mode=mode, x=x_input, y=y_target, s=s.detach().clone())
        acc_cons = get_accuracy(model=model, mode=mode, x=x_input, y=y_target, s=s.detach().clone(), constraint=constraint)

        print(f"Iter {it}: Loss {loss:.4f} | Distortion {distortion:.7f} | Cont. Acc. {acc_unc:.4f} | Disc. Acc. {acc_cons:.4f} | Avg. Sparsity {sparsity:.2f} | L_1 {l_1:.4f}") #| L_inf {l_inf:.4f}"

        n_ex = 4
        fig, ax = plt.subplots(2,n_ex)
        for idx in range(n_ex):
            print()
            ax[0, idx].imshow(x_input[idx].cpu().squeeze(), vmin=0, vmax=1)
            ax[1, idx].imshow(s[idx].cpu().squeeze(), vmin=0, vmax=1)
        fig.set_size_inches(14, 4)
        plt.show()
        
        
class Metrics_collector:
    def __init__(self):
        self.loss = Averager()
        self.merlin_loss = Averager()
        self.morgana_loss = Averager()
        self.merlin_it = Averager()
        self.morgana_it = Averager()
        self.completeness = Averager()
        self.soundness = Averager()
    
    def add(self, loss, merlin_metrics, morgana_metrics):
        self.loss.add(loss)
        self.merlin_loss.add(merlin_metrics['distortion'])
        self.morgana_loss.add(morgana_metrics['distortion'])
        self.merlin_it.add(merlin_metrics['it_until_conv'])
        self.morgana_it.add(morgana_metrics['it_until_conv'])
        self.completeness.add(merlin_metrics['accuracy'])
        self.soundness.add(morgana_metrics['accuracy'])
        
    def add_merlin_only(self, loss, merlin_metrics):
        self.loss.add(loss)
        self.merlin_loss.add(merlin_metrics['distortion'])
        self.merlin_it.add(merlin_metrics['it_until_conv'])
        self.completeness.add(merlin_metrics['accuracy'])
    
    def show_results(self, prefix):
        message = prefix
        message += " loss " + f"{self.loss.result():1.4f}"
        message += " compl. " + f"{100*self.completeness.result():1.2f}%"
        message += " soundn. " + f"{100*self.soundness.result():1.2f}%"
        
        message += " Merlin #it: " + f"{self.merlin_it.result():1.0f}"
        message += " loss: " + f"{self.merlin_loss.result():1.4f}"
        message += "\n"
        message += " Morgana #it: " + f"{self.morgana_it.result():1.0f}"
        message += " loss: " + f"{self.morgana_loss.result():1.4f}"
        print(message)
        
    def return_results(self, prefix):
        message = prefix
        message += " loss " + f"{self.loss.result():1.4f}"
        message += " compl. " + f"{100*self.completeness.result():1.2f}%"
        message += " soundn. " + f"{100*self.soundness.result():1.2f}%"
        
        message += " Merlin #it: " + f"{self.merlin_it.result():1.0f}"
        message += " loss: " + f"{self.merlin_loss.result():1.4f}"
        message += "\n"
        message += " Morgana #it: " + f"{self.morgana_it.result():1.0f}"
        message += " loss: " + f"{self.morgana_loss.result():1.4f}"
        
        return message
        
       
        
def morgana_criterion(output, target):
    device = target.device
    probs = softmax(output).to(device)
    loss1 = (torch.gather(probs,1,target.unsqueeze(1))+probs[:,0]).log().mean()
    return loss1

def merlin_criterion(output, target):
    mc = torch.nn.CrossEntropyLoss().to(target.device)
    return mc(output, target)

def check_stopping(it, max_iter, distortion, dist_old, verbose, stoptol):
    it_crit = (it + 1) > max_iter # Max iterations
    #progress_crit = abs(fk_old - loss.item()) <= stoptol # Progress criterion
    progress_crit = ((1-stoptol)*dist_old - distortion) < 0 # Progress criterion
    stop = it_crit or (progress_crit == 1) # For now, abort when no progress made once
    if verbose and stop:
        #print(it, progress_ctr, fk_old, (1-stoptol)*fk_old, loss.item(), ((1-stoptol)*fk_old - loss.item()))
        print(f"Stopping after it {it}: Iteration criterion {it_crit} | Progress criterion {progress_crit}")
    return stop
