from src.utilities import Averager, categorical_accuracy

import torch

# Routine to calculate test or train loss and accuracy 
def evaluate_model(model, loader, device, lc, nclasses):
    
    mean_loss = Averager()
    mean_accuracy = Averager()
    metrics = dict()
    
    with torch.no_grad():
        for x_input, y_target in loader:
            
            x_image = x_input.select(dim=1, index=0)
            x_image, y_target = x_image.to(device), y_target.to(device)
            
            output = model.eval()(x_image)
            
            loss = lc(output, y_target)
            mean_loss.add(loss.item(), len(y_target))
            mean_accuracy.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
            metrics["loss"] = mean_loss.result()
            metrics["accuracy"] = mean_accuracy.result()
            
    return metrics

def evaluate_model_MC(model, loader, device, lc, nclasses, channels):
    
    mean_loss = Averager()
    mean_accuracy = Averager()
    metrics = dict()
    
    with torch.no_grad():
        for x_input, y_target in loader:
            
            x_image = x_input[:,0:channels,:,:]
            
            x_image, y_target = x_image.to(device), y_target.to(device)
            
            output = model.eval()(x_image)
            
            loss = lc(output, y_target)
            mean_loss.add(loss.item(), len(y_target))
            mean_accuracy.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
            metrics["loss"] = mean_loss.result()
            metrics["accuracy"] = mean_accuracy.result()
            
    return metrics


def regular_training(model, optimizer, nmbr_epochs, trainLoader, testLoader, device, loss_criterion, nclasses):
    
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    
    for epoch in range(nmbr_epochs+1):
        trainLoss_Avg.reset()
        trainAcc_Avg.reset()
        

        if epoch == 0:
            train_metrics = evaluate_model(model=model, loader=trainLoader, device=device, lc = loss_criterion, nclasses=nclasses)
            test_metrics = evaluate_model(model=model, loader=testLoader, device=device, lc = loss_criterion, nclasses=nclasses)
            
            trainLoss, trainAcc = train_metrics["loss"], train_metrics["accuracy"]
            testLoss, testAcc = test_metrics["loss"], test_metrics["accuracy"]
            
        else:
            for x_input, y_target in trainLoader:
                x_image = x_input.select(dim=1, index=0)
               
                x_image, y_target = x_image.to(device), y_target.to(device)

                optimizer.zero_grad()
                output = model.train()(x_image)
                
                loss = loss_criterion(output, y_target)

                loss.backward()
                optimizer.step()
                
                trainLoss_Avg.add(loss.item(), len(y_target))
                trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
            
            trainLoss, trainAcc = trainLoss_Avg.result(), trainAcc_Avg.result()
            test_metrics = evaluate_model(model, loader=testLoader, device=device, lc = loss_criterion, nclasses=nclasses)
            testLoss, testAcc = test_metrics["loss"], test_metrics["accuracy"]
            
        print(f"\nEPOCH {epoch}/{nmbr_epochs}")
        print(f"TRAIN   loss {trainLoss:1.4f} acc {trainAcc*100:2.2f}%")
        print(f"TEST    loss {testLoss:1.4f} acc {testAcc*100:2.2f}%")
        
def regular_training_MC(model, optimizer, nmbr_epochs, trainLoader, testLoader, device, loss_criterion, nclasses, channels):
    
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    
    for epoch in range(nmbr_epochs+1):
        trainLoss_Avg.reset()
        trainAcc_Avg.reset()
        

        if epoch == 0:
            train_metrics = evaluate_model_MC(model=model, loader=trainLoader, device=device, lc = loss_criterion, nclasses=nclasses, channels=channels)
            test_metrics = evaluate_model_MC(model=model, loader=testLoader, device=device, lc = loss_criterion, nclasses=nclasses, channels=channels)
            
            trainLoss, trainAcc = train_metrics["loss"], train_metrics["accuracy"]
            testLoss, testAcc = test_metrics["loss"], test_metrics["accuracy"]
            
        else:
            for x_input, y_target in trainLoader:
                x_image = x_input[:,0:channels,:,:]
               
                x_image, y_target = x_image.to(device), y_target.to(device)

                optimizer.zero_grad()
                output = model.train()(x_image)
                
                
                loss = loss_criterion(output, y_target)

                loss.backward()
                optimizer.step()
                
                trainLoss_Avg.add(loss.item(), len(y_target))
                trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
            
            trainLoss, trainAcc = trainLoss_Avg.result(), trainAcc_Avg.result()
            test_metrics = evaluate_model_MC(model, loader=testLoader, device=device, lc = loss_criterion, nclasses=nclasses,
                                                channels=channels)
            testLoss, testAcc = test_metrics["loss"], test_metrics["accuracy"]
            
        print(f"\nEPOCH {epoch}/{nmbr_epochs}")
        print(f"TRAIN   loss {trainLoss:1.4f} acc {trainAcc*100:2.2f}%")
        print(f"TEST    loss {testLoss:1.4f} acc {testAcc*100:2.2f}%")
        
def evaluate_reg_model(model, loader, params, loss_criterion):
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    device = params['device']
    nclasses = params['nclasses']
    
    for x_input, y_target in loader:
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)

        with torch.no_grad():
            output = model.train()(x_image)
            loss = loss_criterion(output, y_target)

        trainLoss_Avg.add(loss.item(), len(y_target))
        trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
    
    return trainLoss_Avg.result(), trainAcc_Avg.result()
        
        


def train_reg_epoch(model, optimizer, loader, params, loss_criterion):
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    device = params['device']
    nclasses = params['nclasses']
    for x_input, y_target in loader:
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)

        optimizer.zero_grad()
        output = model.train()(x_image)
        loss = loss_criterion(output, y_target)

        loss.backward()
        optimizer.step()
        trainLoss_Avg.add(loss.item(), len(y_target))
        trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
    
    return trainLoss_Avg.result(), trainAcc_Avg.result()

def evaluate_reg_model_MC(model, loader, params, loss_criterion, channels=3):
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    device = params['device']
    nclasses = params['nclasses']
    
    for x_input, y_target in loader:
        x_image = x_input[:,0:channels,:,:]
        x_image, y_target = x_image.to(device), y_target.to(device)

        with torch.no_grad():
            output = model.train()(x_image)
            loss = loss_criterion(output, y_target)

        trainLoss_Avg.add(loss.item(), len(y_target))
        trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
    
    return trainLoss_Avg.result(), trainAcc_Avg.result()
        
        


def train_reg_epoch_MC(model, optimizer, loader, params, loss_criterion, channels=3):
    trainLoss_Avg = Averager()
    trainAcc_Avg = Averager()
    device = params['device']
    nclasses = params['nclasses']
    for x_input, y_target in loader:
        x_image = x_input[:,0:channels,:,:]
        x_image, y_target = x_image.to(device), y_target.to(device)

        optimizer.zero_grad()
        output = model.train()(x_image)
        loss = loss_criterion(output, y_target)

        loss.backward()
        optimizer.step()
        trainLoss_Avg.add(loss.item(), len(y_target))
        trainAcc_Avg.add(categorical_accuracy(y_true=y_target, output=output, nclasses=nclasses), len(y_target))
    
    return trainLoss_Avg.result(), trainAcc_Avg.result()
