from src.models import get_model
from src.stat_evaluation import get_completeness, get_H_cond, fano_violation
import torch
from src.basic_nn_classes import Classifier, SimpleNet
from src.arthur_merlin_networks import evaluate_mask_model
import matplotlib.pyplot as plt
from src.arthur_merlin_utilities import set_labels_data
from torchvision import models

def compute_measures(file_name_fn, trainLoader, testLoader, gamma, masksizeList, use_NN_merlin, use_NN_morgana, params,
                     model_type = 'conv', seed = None):
    #FOR Changing mask sizes

    fixedLoader = [(x,y) for x,y in trainLoader] + [(x,y) for x,y in testLoader]

    H_cond = {} 
    n_ipf = {}
    avg_prec = {}
    P_e = {}
    fano_vio = {}
    

    nclasses = params['nclasses']
    device = params['device']


    for ms in masksizeList:

        print(ms)

        # Get the right model
        model = get_model(model_type=model_type, nclasses=nclasses).to(device)
        trainedmodel_path = file_name_fn("arthur", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
        model.load_state_dict(torch.load(trainedmodel_path))

        merlin_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
        
        # calculate error probability
        if use_NN_merlin == True:
            
            trainedmerlinmodel_path = file_name_fn("merlin", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
            merlin_model.load_state_dict(torch.load(trainedmerlinmodel_path))

            completeness = get_completeness(model=model, loader=testLoader, use_NN=use_NN_merlin, 
                                            merlin_model=merlin_model, mask_size=ms, params = params)
        else:
            completeness = get_completeness(model=model, loader=testLoader, use_NN=use_NN_merlin, mask_size=ms, 
                                            params = params)
        P_e[ms] = 1 - completeness

        # Calculate conditional Entropy
        avg_prec[ms], H_cond[ms], n_ipf[ms] = get_H_cond(model=model, testLoader = testLoader, fixedLoader = fixedLoader, 
                                                                  mask_size=ms, diffTol = 1E-7, params=params, 
                                                                  merlin_model=merlin_model,use_NN=use_NN_merlin)

        # Calculate Fano violation
        fano_vio[ms] = fano_violation(P_e[ms], H_cond[ms],nclasses)
    
    return P_e, fano_vio, avg_prec, H_cond, n_ipf


def precision_bound(completeness, soundness, afc=1, class_imbalance=1):
    
    eps_c = 1-completeness
    eps_s = 1-soundness
    
    return 1 - eps_c - afc*eps_s/(1 - eps_c + afc*eps_s*class_imbalance)

def eval_soundness_completeness(file_name_fn, masksizeList, testLoader, gamma, use_NN_merlin, use_NN_morgana, params,
                                model_type='conv', seed= None):
        # Init model
        
        
    avg_prec_bnd = {}
    completeness = {}
    soundness = {}
    
    device = params['device']
    nclasses = params['nclasses']
    
    model = get_model(model_type=model_type, nclasses=nclasses).to(device)

    #Select Optimizer
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)

    merlin_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
    merlin_model_opt = torch.optim.Adam(merlin_model.parameters(), lr=1e-4)

    morgana_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
    morgana_model_opt = torch.optim.Adam(morgana_model.parameters(), lr=1e-6)

    for ms in masksizeList:

        trainedmodel_path = file_name_fn("arthur", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
        trained_merlin_model_path = file_name_fn("merlin", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
        trained_morgana_model_path = file_name_fn("morgana", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)


        model.load_state_dict(torch.load(trainedmodel_path))
        
        if use_NN_merlin == True:
            merlin_model.load_state_dict(torch.load(trained_merlin_model_path))
        else:
            merlin_model = None

        if use_NN_morgana == True:
            morgana_model.load_state_dict(torch.load(trained_morgana_model_path))
        else:
            morgana_model = None

        test_collector = evaluate_mask_model(model, use_NN_merlin, use_NN_morgana, merlin_model, merlin_model_opt, 
                                             morgana_model, morgana_model_opt, loader=testLoader, gamma=gamma, 
                                             mask_size=ms, max_iter=0, params=params)

        soundness[ms] = test_collector.soundness.result()
        completeness[ms] = test_collector.completeness.result()

        test_collector.show_results("TEST   ")

    
    for k in soundness.keys():
        avg_prec_bnd[k] = precision_bound(completeness[k], soundness[k])
    
    return completeness, soundness, avg_prec_bnd

def eval_soundness_completeness_class_specific(file_name_fn, masksizeList, testData, gamma, use_NN_merlin, 
                                               use_NN_morgana, params, model_type='conv', seed = None):
        # Init model
        
        
    avg_prec_bnd = {}
    completeness = {}
    soundness = {}
    
    device = params['device']
    nclasses = params['nclasses']
    classes = list(range(1,nclasses))
    
    model = get_model(model_type=model_type, nclasses=nclasses).to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)

    merlin_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
    merlin_model_opt = torch.optim.Adam(merlin_model.parameters(), lr=1e-4)

    morgana_model = SimpleNet(nchannels=1,nclasses=nclasses).to(device)
    morgana_model_opt = torch.optim.Adam(morgana_model.parameters(), lr=1e-6)
    
    

    for ms in masksizeList:
    
        trainedmodel_path = file_name_fn("arthur", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
        trained_merlin_model_path = file_name_fn("merlin", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)
        trained_morgana_model_path = file_name_fn("morgana", model_type, gamma, ms, use_NN_merlin, use_NN_morgana, seed)


        print(trainedmodel_path)
        model.load_state_dict(torch.load(trainedmodel_path))
        
        if use_NN_merlin == True:
            merlin_model.load_state_dict(torch.load(trained_merlin_model_path))
        else:
            merlin_model = None

        if use_NN_morgana == True:
            morgana_model.load_state_dict(torch.load(trained_morgana_model_path))
        else:
            morgana_model = None

        soundness[ms] = {}
        completeness[ms] = {}
        
        print(f"size of dataset:{len(testData)}")
        
        for cls in classes:
            print(f"mask size:{ms} and class:{cls}")
            testDatacls = set_labels_data(testData, [cls])
            print(f"size of dataset:{len(testDatacls)}")
                                     
            testLoadercls = torch.utils.data.DataLoader(testDatacls, batch_size=128, shuffle=False,
                                                                  pin_memory=torch.cuda.is_available(), num_workers=2)
            
            test_collector = evaluate_mask_model(model, use_NN_merlin, use_NN_morgana, merlin_model, merlin_model_opt, 
                                                 morgana_model, morgana_model_opt, loader=testLoadercls, gamma=gamma, 
                                                 mask_size=ms, max_iter=0, params=params)

            soundness[ms][cls] = test_collector.soundness.result()
            completeness[ms][cls] = test_collector.completeness.result()

            test_collector.show_results("TEST   ")

    
    return completeness, soundness

def plot_parameters(soundness, completeness, avg_prec, avg_prec_bnd, filename):
    fig, ax1 = plt.subplots(figsize=(15,12))

    from math import log2
    xvals = [log2(i) for i in soundness.keys()]
    ax1.plot(xvals, soundness.values(),label="Soundness", linewidth=4, linestyle=':',color='r')
    ax1.plot(xvals, completeness.values(),label="Completeness", linewidth=4, linestyle='-.',color='b')
    ax1.plot(xvals, avg_prec.values(),label="Average Precision",color='k', linewidth=4, linestyle='--')
    ax1.plot(xvals, avg_prec_bnd.values(),label="Avg Prec Bnd",color='k', linewidth=4,linestyle='-')
    ax1.set_title("Merlin: NN, Morgana:NN")
    ax1.set_xlabel("Mask size (log2)")
    ax1.set_ylim(-0.15,1.1)
    ax1.legend(loc='lower right')

    fig.tight_layout()
    plt.savefig(filename, transparent=True)
    
    
def plot_parameters_detailed(soundness, completeness, avg_prec, avg_prec_bnd, start_plot=3):
    fig, ax1 = plt.subplots()
    ax1.plot(list(soundness.keys())[start_plot:], list(soundness.values())[start_plot:],label="Soundness")
    ax1.plot(list(completeness.keys())[start_plot:], list(completeness.values())[start_plot:],label="Completeness")
    ax1.plot(list(avg_prec.keys())[start_plot:], list(avg_prec.values())[start_plot:],label="Average Precision",color='k')
    ax1.plot(list(avg_prec_bnd.keys())[start_plot:], list(avg_prec_bnd.values())[start_plot:],label="Avg Prec Bnd",color='b')


    ax1.legend(loc='lower right')
    fig.tight_layout()