import torch
import numpy as np

def set_labels(trainData, testData, classes):
    label_count = 1
    train_idxs = torch.zeros_like(trainData.targets)
    test_idxs = torch.zeros_like(testData.targets)
    
    for label in classes:    
        train_idxs = torch.logical_or(train_idxs, trainData.targets==label)
        test_idxs = torch.logical_or(test_idxs, testData.targets==label)
        
        #To since the label of 0 is always assigned to the I don't know option, 
        #we assing the label of 10 to images with the digit 0
        if label == 0:
            trainData.targets[trainData.targets==label] = len(classes) + 1
            testData.targets[testData.targets==label] = len(classes) + 1
            label_count += 1
        else:
            trainData.targets[trainData.targets==label] = label_count
            testData.targets[testData.targets==label] = label_count
            label_count += 1
            
    trainData.data = trainData.data[train_idxs]
    trainData.targets = trainData.targets[train_idxs]
    testData.data = testData.data[test_idxs]
    testData.targets = testData.targets[test_idxs]
    
    return trainData, testData

def get_label_loader(trainData, classes):
    label_count = 1
    train_idxs = torch.zeros_like(trainData.targets)
    
    for label in classes:    
        train_idxs = torch.logical_or(train_idxs, trainData.targets==label)
    trainData.data = trainData.data[train_idxs]
    trainData.targets = trainData.targets[train_idxs]
    
    return trainData

def get_altered_dataset(OriginalDataset):
    # This only works for MNIST for now!
    class AlteredDatasetWrapper(OriginalDataset):
        """Note: this will download the dataset again since the class name changed."""

        def __init__(self, *args, **kwargs):
            super(AlteredDatasetWrapper, self).__init__(*args, **kwargs)
            self.masks = torch.empty_like(self.data, dtype=torch.float32).uniform_()
            #print(self.masks.shape)
            
        def update_mask(self, index, new_mask):
            """
            index: tensor of indices
            new_mask: new_mask in shape (len(indices), 28, 28)
            """
            self.masks[index] = new_mask

        def __getitem__(self, index):
            # Overload this to jointly return image and mask
            image, label = super(AlteredDatasetWrapper, self).__getitem__(index=index)
            mask = self.masks[index]
            #joined = torch.cat((image, mask.view(image.shape)), dim=0)
            return image, mask, label, index
    return AlteredDatasetWrapper


class AddMaskChannel(object):
    """Convert ndarrays in sample to Tensors."""
    def __init__(self, init='uniform'):
        self.init = init
        
    
    def __call__(self, sample):
        # Add a mask value for every pixel
        # Important: This does not work when using more than 1 channel, since the mask will be too large then
        if self.init != 'uniform':
            sample = torch.cat((sample, torch.empty_like(sample).fill_(self.init)))
        else:
            sample = torch.cat((sample, torch.empty_like(sample).uniform_()))
        return sample

class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.sum = 0
        self.count = 0
        self.avg = 0

    def result(self):
        if self.count > 0:
            return self.avg
        else:
            return np.nan

    def __call__(self, val, n=1):
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        if self.count > 0:
            return str(self.avg)
        else:
            return np.nan
        
        
class Averager(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.sum = 0
        self.count = 0
        self.avg = 0

    def result(self):
        if self.count > 0:
            return self.avg
        else:
            return np.nan

    def add(self, val, n=1):
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
        
@torch.no_grad()
def categorical_accuracy(y_true, output, nclasses):
    if nclasses==2:
        prediction = (output>0).type_as(y_true).reshape(y_true.shape)
        return prediction.eq(y_true).sum().item() / float(len(y_true))
    else:
        prediction = torch.argmax(output, dim=1)
        return prediction.eq(y_true).sum().item() / float(len(y_true))

