import torch as th
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from custom_criterion import custom_criterion
from mask_optimization import brute_force_mask, opt_mask
from metrics import get_accuracy, get_soundness, print_metrics
from utilities import Averager, plot_to_tensorboard, seed_worker


def adversarial_train_epoch(
    model, optimizer, train_data, device, k=1, gamma=0.75, alpha=0, brute_force=False
):
    """_summary_

    Args:
        model (torch.nn.model): Classifier
        optimizer (torch.optim): Model optimizer (e.g., Adam or SGD so far).
        train_data (torch.utils.data): Training dataset containing the datapoints and masks.
        device (str): Cuda or CPU.
        k (int, optional): Size of mask. Defaults to 1.
        gamma (float, optional): Weight parameter of Merlins and Morganas loss in Arthurs optimization. Defaults to 0.75.
        alpha (float, optional): Factor for penalty.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        metrics (dict): Dictionary with metrics.
    """
    output_size = model.output_size
    train_loss = Averager()
    train_completeness = Averager()
    train_soundness = Averager()
    metrics = {}
    merlin_criterion = th.nn.CrossEntropyLoss()
    morgana_criterion = custom_criterion
    # Initialize dataloader with updated masks
    train_dataloader = DataLoader(
        train_data,
        batch_size=512,
        shuffle=True,
        num_workers=4,
    )
    for y_target, x_input, mask_merlin, mask_morgana, idx in tqdm(train_dataloader):
        y_target = y_target.to(device)
        x_input = x_input.to(device)
        mask_merlin = mask_merlin.to(device)
        mask_morgana = mask_morgana.to(device)

        # Optimize with Merlin
        x_merlin, mask_merlin, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="merlin",
            k=k,
            mask=mask_merlin,
            device=device,
            lr=1e-1,
            alpha=alpha,
        )

        if brute_force:
            # Perfect Morgana
            x_morgana, mask_morgana = brute_force_mask(
                model,
                x_input,
                y_target,
                mode="morgana",
                k=k,
                device=device,
            )
        else:
            # Optimize with Morgana
            x_morgana, mask_morgana, _ = opt_mask(
                model,
                x_input,
                y_target,
                mode="morgana",
                k=k,
                mask=mask_morgana,
                device=device,
                lr=1e-1,
                verbose=False,
            )

        # update new mask to train dataset
        train_data.update_mask(mask_merlin.to("cpu"), mode="merlin", idx=idx)
        train_data.update_mask(mask_morgana.to("cpu"), mode="morgana", idx=idx)

        optimizer.zero_grad()
        logits_merlin = model.train()(x_merlin)
        logits_morgana = model.train()(x_morgana)

        merlin_loss = merlin_criterion(logits_merlin, y_target)
        morgana_loss = morgana_criterion(logits_morgana, y_target)
        loss = merlin_loss - gamma * morgana_loss
        loss.backward()
        th.nn.utils.clip_grad_norm_(model.parameters(), 1)
        optimizer.step()

        # calculate metrics
        completeness = get_accuracy(logits_merlin, y_target, output_size=output_size)
        soundness = get_soundness(logits_morgana, y_target)
        # log epoch results
        train_loss.add(loss.item(), len(y_target))
        train_completeness.add(completeness, len(y_target))
        train_soundness.add(soundness, len(y_target))

    metrics["loss"] = train_loss.result()
    metrics["completeness"] = train_completeness.result()
    metrics["soundness"] = train_soundness.result()

    return metrics


def adversarial_eval_epoch(
    model, test_dataloader, logger, device, epoch, k=1, gamma=0.5, alpha=0, brute_force=False
):
    """Tests Arthur with Merlin and Morgana incorporated.

    Args:
        model (torch.nn.model): Classifier
        test_dataloader (torch.utils.data): Test dataloader containing the datapoints and masks.
        logger (torch.nn.utils): Logger (Tensorboard).
        device (str): Cuda or CPU.
        epoch (int): Current epoch.
        k (int, optional): Size of mask. Defaults to 1.
        gamma (float, optional): Weight parameter of Merlins and Morganas loss in Arthurs optimization. Defaults to 0.75.
        alpha (float, optional): Factor for penalty. Defaults to 0.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        metrics (dict): Dictionary with metrics.
    """
    output_size = model.output_size
    test_loss = Averager()
    test_completeness = Averager()
    test_soundness = Averager()
    metrics = {}
    n_features, _ = test_dataloader.dataset[0][1].shape
    cum_merlin_masks = th.zeros([1, n_features]).to(device)
    cum_morgana_masks = th.zeros([1, n_features]).to(device)
    merlin_criterion = th.nn.CrossEntropyLoss()
    morgana_criterion = custom_criterion
    # Initialize dataloader with updated masks
    for y_target, x_input, mask_merlin, mask_morgana, idx in tqdm(test_dataloader):
        y_target = y_target.to(device)
        x_input = x_input.to(device)
        mask_merlin = mask_merlin.to(device)
        mask_morgana = mask_morgana.to(device)

        # Optimize with Merlin
        x_merlin, mask_merlin, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="merlin",
            k=k,
            mask=mask_merlin,
            device=device,
            lr=1e-1,
            alpha=alpha,
        )

        if brute_force:
            # Perfect Morgana
            x_morgana, mask_morgana = brute_force_mask(
                model,
                x_input,
                y_target,
                mode="morgana",
                k=k,
                device=device,
            )
        else:
            # Optimize with Morgana
            x_morgana, mask_morgana, _ = opt_mask(
                model,
                x_input,
                y_target,
                mode="morgana",
                k=k,
                mask=mask_morgana,
                device=device,
                lr=1e-1,
            )

        with th.no_grad():
            cum_merlin_masks += mask_merlin.sum(dim=0)
            cum_morgana_masks += mask_morgana.sum(dim=0)

            logits_merlin = model.eval()(x_merlin)
            logits_morgana = model.eval()(x_morgana)

            merlin_loss = merlin_criterion(logits_merlin, y_target)
            morgana_loss = morgana_criterion(logits_morgana, y_target)
            loss = merlin_loss - gamma * morgana_loss
            # calculate metrics
            completeness = get_accuracy(logits_merlin, y_target, output_size=output_size)
            soundness = get_soundness(logits_morgana, y_target)
            # log epoch results
            test_loss.add(loss.item(), len(y_target))
            test_completeness.add(completeness, len(y_target))
            test_soundness.add(soundness, len(y_target))

    metrics["loss"] = test_loss.result()
    metrics["completeness"] = test_completeness.result()
    metrics["soundness"] = test_soundness.result()
    merlin_feat_frac = cum_merlin_masks.to("cpu") / (k * len(test_dataloader.dataset))
    morgana_feat_frac = cum_morgana_masks.to("cpu") / (k * len(test_dataloader.dataset))
    feature_list = list(test_dataloader.dataset.dataset.columns)

    plot_to_tensorboard(
        logger, merlin_feat_frac, morgana_feat_frac, metrics, feature_list, gamma, alpha, epoch
    )

    logger.flush()
    return metrics


def adversarial_training(
    model,
    train_data,
    test_dataloader,
    device,
    mask_size,
    nepochs,
    lr=1e-4,
    gamma=0.75,
    alpha=0,
    brute_force=False,
):
    """Trains the classifier with cooperative Merlin and adversarial Morgana,

    which are optimized via SFW, for multiple epochs.

    Args:
        model (torch.nn.model): Classifier.
        train_data (torch.utils.data): Training dataset containing the datapoints and masks.
        test_data (torch.utils.data): Test dataset containing the datapoints and masks.
        device (str): Cuda or CPU.
        mask_size (int): Size of mask.
        nepochs (int): Number of epochs per mask size.
        lr (float, optional): Learning rate of model optimizer. Defaults to 0.01.
        gamma (int, optional): Weight on Morgana's loss.
        alpha (float, optional): Factor for penalty.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        No return
    """
    logger = SummaryWriter(comment=f"_gamma{gamma}_alpha{int(alpha)}_k{mask_size}_lr{lr:1.1E}")
    optimizer = th.optim.SGD(model.parameters(), lr=lr)
    soundness_threshold = 0.9
    best_completeness = 0
    for epoch in range(0, nepochs + 1):
        print(20 * "---")
        print(f"Epoch: {epoch} | Train with mask size: {mask_size}")
        train_metrics = adversarial_train_epoch(
            model,
            optimizer,
            train_data,
            device,
            k=mask_size,
            gamma=gamma,
            alpha=alpha,
            brute_force=brute_force,
        )
        test_metrics = adversarial_eval_epoch(
            model,
            test_dataloader,
            logger,
            device,
            epoch,
            k=mask_size,
            gamma=gamma,
            alpha=alpha,
            brute_force=brute_force,
        )
        # print results
        print_metrics("Train ", train_metrics)
        print_metrics("Test ", test_metrics)

        if gamma > 0:
            if test_metrics["soundness"] > soundness_threshold: 
                    if test_metrics["completeness"] >= best_completeness:
                        best_completeness = test_metrics["completeness"]
                        th.save(
                            dict(
                                model=model.state_dict(),
                                learning_rate=lr,
                                optimizer=optimizer.state_dict(),
                                completeness=best_completeness,
                                soundness=test_metrics["soundness"],
                                gamma=gamma,
                                alpha=alpha,
                                epoch=epoch,
                            ),
                            f"{logger.log_dir}/best_checkpoint_epoch.pth",
                        )
        elif gamma == 0:
            if test_metrics["completeness"] >= best_completeness:
                best_completeness = test_metrics["completeness"]
                th.save(
                    dict(
                        model=model.state_dict(),
                        learning_rate=lr,
                        optimizer=optimizer.state_dict(),
                        completeness=best_completeness,
                        gamma=gamma,
                        alpha=alpha,
                        epoch=epoch,
                    ),
                    f"{logger.log_dir}/best_checkpoint_epoch.pth",
                )

    logger.close()


def train_comparison(
    model, optimizer, train_data, device, k=1, gamma=0.75, alpha=0, brute_force=False
):
    """Trains Arthur with Merlin and Morgana as optimizers and compares soundness to perfect Morgana.

    Does not train with masks obtained from brute force.

    Args:
        model (torch.nn.model): Classifier
        optimizer (torch.optim): Model optimizer (e.g., Adam or SGD so far).
        train_data (torch.utils.data): Training dataset containing the datapoints and masks.
        device (str): Cuda or CPU.
        k (int, optional): Size of mask. Defaults to 1.
        gamma (float, optional): Weight parameter of Merlins and Morganas loss in Arthurs optimization. Defaults to 0.75.
        alpha (float, optional): Factor for penalty.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        metrics (dict): Dictionary with metrics.
    """
    output_size = model.output_size
    train_loss = Averager()
    train_completeness = Averager()
    train_soundness = Averager()
    train_soundness_bf = Averager()
    metrics = {}
    merlin_criterion = th.nn.CrossEntropyLoss()
    morgana_criterion = custom_criterion
    train_dataloader = DataLoader(
        train_data,
        batch_size=512,
        shuffle=True,
        num_workers=4,
    )
    for y_target, x_input, mask_merlin, mask_morgana, idx in tqdm(train_dataloader):
        y_target = y_target.to(device)
        x_input = x_input.to(device)
        mask_merlin = mask_merlin.to(device)
        mask_morgana = mask_morgana.to(device)

        # Optimize with Merlin
        x_merlin, mask_merlin, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="merlin",
            k=k,
            mask=mask_merlin,
            device=device,
            lr=1e-1,
            alpha=alpha,
        )
        # Optimize with Morgana
        x_morgana, mask_morgana, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="morgana",
            k=k,
            mask=mask_morgana,
            device=device,
            lr=1e-1,
            verbose=False,
        )
        # Perfect Morgana
        x_masked_bf, mask_bf = brute_force_mask(
            model,
            x_input,
            y_target,
            mode="morgana",
            k=k,
            device=device,
        )

        # update new mask to train dataset
        train_data.update_mask(mask_merlin.to("cpu"), mode="merlin", idx=idx)
        train_data.update_mask(mask_morgana.to("cpu"), mode="morgana", idx=idx)

        optimizer.zero_grad()
        logits_merlin = model.train()(x_merlin)
        logits_morgana = model.train()(x_morgana)
        logits_bf = model.eval()(x_masked_bf)

        merlin_loss = merlin_criterion(logits_merlin, y_target)
        morgana_loss = morgana_criterion(logits_morgana, y_target)
        loss = merlin_loss - gamma * morgana_loss
        loss.backward()
        th.nn.utils.clip_grad_norm_(model.parameters(), 1)
        optimizer.step()

        # calculate metrics
        completeness = get_accuracy(logits_merlin, y_target, output_size=output_size)
        soundness = get_soundness(logits_morgana, y_target)
        soundness_bf = get_soundness(logits_bf, y_target)

        # log epoch results
        train_loss.add(loss.item(), len(y_target))
        train_completeness.add(completeness, len(y_target))
        train_soundness.add(soundness, len(y_target))
        train_soundness_bf.add(soundness_bf, len(y_target))

    metrics["loss"] = train_loss.result()
    metrics["completeness"] = train_completeness.result()
    metrics["soundness"] = train_soundness.result()
    metrics["soundness_bf"] = train_soundness_bf.result()

    return metrics


def test_comparison(
    model, test_dataloader, logger, device, epoch, k=1, gamma=0.5, alpha=0, brute_force=False
):
    """Tests Arthur with Merlin and Morgana incorporated.

    Soundness is compared to soundness achieved with respect to perfect Morgana.

    Args:
        model (torch.nn.model): Classifier
        test_dataloader (torch.utils.data): Test dataloader containing the datapoints and masks.
        logger (torch.nn.utils): Logger (Tensorboard).
        device (str): Cuda or CPU.
        epoch (int): Current epoch.
        k (int, optional): Size of mask. Defaults to 1.
        gamma (float, optional): Weight parameter of Merlins and Morganas loss in Arthurs optimization. Defaults to 0.75.
        alpha (float, optional): Factor for penalty. Defaults to 0.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        metrics (dict): Dictionary with metrics.
    """
    output_size = model.output_size

    test_loss = Averager()
    test_loss_bf = Averager()
    test_completeness = Averager()
    test_soundness = Averager()
    test_soundness_bf = Averager()
    metrics = {}

    n_features, _ = test_dataloader.dataset[0][1].shape

    cum_merlin_masks = th.zeros([1, n_features]).to(device)
    cum_morgana_masks = th.zeros([1, n_features]).to(device)
    cum_bf_masks = th.zeros([1, n_features]).to(device)

    merlin_criterion = th.nn.CrossEntropyLoss()
    morgana_criterion = custom_criterion
    for y_target, x_input, mask_merlin, mask_morgana, idx in tqdm(test_dataloader):
        y_target = y_target.to(device)
        x_input = x_input.to(device)
        mask_merlin = mask_merlin.to(device)
        mask_morgana = mask_morgana.to(device)

        # Optimize with Merlin
        x_merlin, mask_merlin, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="merlin",
            k=k,
            mask=mask_merlin,
            device=device,
            lr=1e-1,
            alpha=alpha,
        )
        # Optimize with Morgana
        x_morgana, mask_morgana, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="morgana",
            k=k,
            mask=mask_morgana,
            device=device,
            lr=1e-1,
            verbose=False,
        )
        # Perfect Morgana
        x_masked_bf, mask_bf = brute_force_mask(
            model,
            x_input,
            y_target,
            mode="morgana",
            k=k,
            device=device,
        )

        with th.no_grad():
            cum_merlin_masks += mask_merlin.sum(dim=0)
            cum_morgana_masks += mask_morgana.sum(dim=0)
            cum_bf_masks += mask_bf.sum(dim=0)

            logits_merlin = model.eval()(x_merlin)
            logits_morgana = model.eval()(x_morgana)
            logits_bf = model.eval()(x_masked_bf)

            merlin_loss = merlin_criterion(logits_merlin, y_target)
            morgana_loss = morgana_criterion(logits_morgana, y_target)
            bf_morgana_loss = morgana_criterion(logits_bf, y_target)
            loss = merlin_loss - gamma * morgana_loss
            loss_bf = merlin_loss - gamma * bf_morgana_loss

            # calculate metrics
            completeness = get_accuracy(logits_merlin, y_target, output_size=output_size)
            soundness = get_soundness(logits_morgana, y_target)
            soundness_bf = get_soundness(logits_bf, y_target)

            # log epoch results
            test_loss.add(loss.item(), len(y_target))
            test_loss_bf.add(loss_bf.item(), len(y_target))
            test_completeness.add(completeness, len(y_target))
            test_soundness.add(soundness, len(y_target))
            test_soundness_bf.add(soundness_bf, len(y_target))

    metrics["loss"] = test_loss.result()
    metrics['loss_bf'] = test_loss_bf.result()
    metrics["completeness"] = test_completeness.result()
    metrics["soundness"] = test_soundness.result()
    metrics["soundness_bf"] = test_soundness_bf.result()


    merlin_feat_frac = cum_merlin_masks.to("cpu") / (k * len(test_dataloader.dataset))
    morgana_feat_frac = cum_morgana_masks.to("cpu") / (k * len(test_dataloader.dataset))
    brute_force_feat_frac = cum_bf_masks.to("cpu") / (k * len(test_dataloader.dataset))
    feature_list = list(test_dataloader.dataset.dataset.columns)

    plot_to_tensorboard(
        logger,
        merlin_feat_frac,
        morgana_feat_frac,
        metrics,
        feature_list,
        gamma,
        alpha,
        epoch,
    )
    metrics_bf = {'completeness': metrics['completeness'], 'soundness': metrics['soundness_bf']}
    plot_to_tensorboard(
        logger,
        merlin_feat_frac,
        brute_force_feat_frac,
        metrics_bf,
        feature_list,
        gamma,
        alpha,
        epoch,
        tag="Brute Force Selection"
    )

    logger.flush()

    return metrics


def compare_to_brute_force(
    model,
    train_data,
    test_dataloader,
    device,
    mask_size,
    nepochs,
    lr=1e-4,
    gamma=0.75,
    alpha=0,
    brute_force=False,
):
    """Trains the classifier with cooperative Merlin and adversarial Morgana,

    which are optimized via SFW, for multiple epochs.

    Args:
        model (torch.nn.model): Classifier.
        train_data (torch.utils.data): Training dataset containing the datapoints and masks.
        test_data (torch.utils.data): Test dataset containing the datapoints and masks.
        device (str): Cuda or CPU.
        mask_size (int): Size of mask.
        nepochs (int): Number of epochs per mask size.
        lr (float, optional): Learning rate of model optimizer. Defaults to 0.01.
        gamma (int, optional): Weight on Morgana's loss.
        alpha (float, optional): Factor for penalty.
        brute_force (bool, optional): Lets Morgana use the brute-force-method. Defaults to False.

    Returns:
        No return
    """
    logger = SummaryWriter(comment=f"_gamma{gamma}_alpha{int(alpha)}_k{mask_size}_lr{lr:1.1E}_bruteforce_{brute_force}")
    optimizer = th.optim.Adam(model.parameters(), lr=lr)
    best_avg_score = 0
    for epoch in range(0, nepochs + 1):
        print(20 * "---")
        print(f"Epoch: {epoch} | Train with mask size: {mask_size}")
        train_metrics = train_comparison(
            model,
            optimizer,
            train_data,
            device,
            k=mask_size,
            gamma=gamma,
            alpha=alpha,
            brute_force=brute_force,
        )
        test_metrics = test_comparison(
            model,
            test_dataloader,
            logger,
            device,
            epoch,
            k=mask_size,
            gamma=gamma,
            alpha=alpha,
            brute_force=brute_force,
        )
        # print results
        print_metrics("Train ", train_metrics)
        print_metrics("Test ", test_metrics)

        if gamma > 0:
            if (test_metrics["completeness"] + test_metrics["soundness"]) / 2 >= best_avg_score:
                best_completeness = test_metrics["completeness"]
                best_soundness = test_metrics["soundness"]
                best_avg_score = (best_completeness + best_soundness) / 2
                th.save(
                    dict(
                        model=model.state_dict(),
                        learning_rate=lr,
                        optimizer=optimizer.state_dict(),
                        completeness=best_completeness,
                        soundness=best_soundness,
                        soundness_bf=test_metrics["soundness_bf"],
                        gamma=gamma,
                        alpha=alpha,
                        epoch=epoch,
                    ),
                    f"{logger.log_dir}/best_checkpoint_epoch{epoch}.pth",
                )
        elif gamma == 0:
            if test_metrics["completeness"] >= best_avg_score:
                best_completeness = test_metrics["completeness"]
                best_avg_score = best_completeness
                th.save(
                    dict(
                        model=model.state_dict(),
                        learning_rate=lr,
                        optimizer=optimizer.state_dict(),
                        completeness=best_completeness,
                        gamma=gamma,
                        alpha=alpha,
                        epoch=epoch,
                    ),
                    f"{logger.log_dir}/best_checkpoint_epoch{epoch}.pth",
                )

    logger.close()
