import glob

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from census_data import CensusIncomeDataset, read_census_data
from mask_optimization import opt_mask
from metrics import get_accuracy, get_soundness
from model_training import ArthurClassifier
from utilities import Averager

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print("Running on Device:", device)

torch.manual_seed(42)

params = {
    "text.usetex": True,
    "font.size": 18,
    "font.family": "serif",
    "font.serif": "Times New Roman",
}
plt.rcParams.update(params)
# Read dataframe into memory
train_data, test_data = read_census_data(PATH="data/adult", target_class="sex_target")
# Prepare train/test dataloader
test_data = CensusIncomeDataset(dataset=test_data, target_class="sex_target")
test_dataloader = DataLoader(test_data, batch_size=512, shuffle=False, num_workers=4)
n_features, encoding_size = test_data[0][1].shape

model = ArthurClassifier(input_size=n_features * encoding_size, output_size=3).to(device)

model_paths = glob.glob('runs/*/*.pth')
for path in model_paths:
    current_model = torch.load(path)
    model.load_state_dict(current_model["model"])
    gamma = current_model["gamma"]
    alpha = current_model["alpha"]
    k = 1

    test_completeness = Averager()
    test_soundness = Averager()
    cum_merlin_masks = torch.zeros([1, n_features]).to(device)
    cum_morgana_masks = torch.zeros([1, n_features]).to(device)

    cum_merlin_idk = torch.zeros([1, n_features]).to(device)
    cum_morgana_idk = torch.zeros([1, n_features]).to(device)
    for y_target, x_input, mask_merlin, mask_morgana, idx in tqdm(test_dataloader):
        y_target = y_target.to(device)
        x_input = x_input.to(device)
        mask_merlin = mask_merlin.to(device)
        mask_morgana = mask_morgana.to(device)
        x_merlin, mask_merlin, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="merlin",
            k=k,
            mask=mask_merlin,
            device=device,
            verbose=False,
            alpha=alpha,
        )
        # x_input is being masked with k-sparse mask
        x_morgana, mask_morgana, _ = opt_mask(
            model,
            x_input,
            y_target,
            mode="morgana",
            k=k,
            mask=mask_morgana,
            device=device,
            verbose=False,
        )

        with torch.no_grad():
            cum_merlin_masks += mask_merlin.sum(dim=0)
            cum_morgana_masks += mask_morgana.sum(dim=0)

            logits_merlin = model.eval()(x_merlin)
            logits_morgana = model.eval()(x_morgana)

            cum_merlin_idk += mask_merlin[torch.argmax(logits_merlin, dim=1) == 2].sum(dim=0)
            cum_morgana_idk += mask_morgana[torch.argmax(logits_morgana, dim=1) == 2].sum(dim=0)
            # calculate metrics
            completeness = get_accuracy(logits_merlin, y_target, output_size=model.output_size)
            soundness = get_soundness(logits_morgana, y_target)
            test_completeness.add(completeness, len(y_target))
            test_soundness.add(soundness, len(y_target))

    final_completeness = test_completeness.result()
    final_soundness = test_soundness.result()

    merlin_idk_frac = cum_merlin_idk.to("cpu") / len(test_data)
    morgana_idk_frac = cum_morgana_idk.to("cpu") / len(test_data)
    merlin_feat_frac = cum_merlin_masks.to("cpu") / len(test_data)
    morgana_feat_frac = cum_morgana_masks.to("cpu") / len(test_data)

    print(f"Soundness: {final_soundness:1.3}")
    print(f"Completeness: {final_completeness}")
    print(f"Merlin features: {merlin_feat_frac}")
    print(f"Morgana features: {morgana_feat_frac}")

    feature_list = list(test_data.dataset.columns)
    feature_list[2] = "education"  # replace by proper name

    width = 0.4
    if gamma == 0:
        fig, ax = plt.subplots(figsize=(6.4, 3.62))
    else:
        fig, ax = plt.subplots(figsize=(6.4, 4.8))
    # # You must select the correct size of the plot in advance
    # fig.set_size_inches(2.5,2.5)
    textstr = "\n".join(
        (
            r"$Completeness=%.0f\%%$" % (100 * final_completeness,),
            r"$Soundness=%.2f\%%$" % (100 * final_soundness,) if gamma > 0 else "",
        )
    )

    # place a text in upper left in axes coords
    ax.text(0.01, 0.98, textstr, transform=ax.transAxes, verticalalignment="top")
    xt = np.arange(len(feature_list))
    ax.set_ylim(bottom=0, top=1)

    # Plot Merlin's selection
    ax.bar(
        xt - width / 2,
        merlin_feat_frac.flatten(),
        width,
        label="Chosen by Merlin",
        color="mediumturquoise",  # "mediumturquoise" for less saturation | "blue" for more saturation
    )

    if gamma > 0:
        # Plot Morgana's selection
        ax.bar(
            xt + width / 2,
            morgana_feat_frac.flatten(),
            width,
            label="Chosen by Morgana",
            color="salmon",  # "salmon" for less saturation | "red" for more saturation
        )
    else:
        ax.bar(
        xt + width / 2,
        0. * morgana_feat_frac.flatten(),
        width,
        label="Chosen by Morgana",
        color="salmon",  # "salmon" for less saturation | "red" for more saturation
    )

    if gamma == 0:
        ax.set_xticks(xt, [])
    else:
        ax.set_xticks(xt, feature_list)
    ax.tick_params(axis="x", rotation=45)
    plt.setp(ax.xaxis.get_majorticklabels(), ha="right", rotation_mode="anchor")
    # ax.set_xlabel("Feature")
    ax.set_ylabel("Relative Occurence")
    # ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.8), ncol=2)
    ax.grid(axis="y", alpha=0.3)
    fig.tight_layout()

    print(f"size:{fig.get_size_inches()}")
    print(f"dpi:{fig.dpi}")
    plt.savefig(f"plot_penalty{alpha}_gamma{gamma}.pdf", transparent=True, edgecolor="none")

