import os

import numpy as np
import pandas as pd
import sh
import torch as th
import wget
from imblearn.under_sampling import RandomUnderSampler
from sklearn.preprocessing import LabelBinarizer, MinMaxScaler, OneHotEncoder
from torch.utils.data import Dataset

# Comments are about number of NaN in training set
header_names = [
    "age",  # clean
    "workclass",  # 1836 NaN out of 32560 samples
    "fnlwgt",  # being dropped
    "education",  # clean
    "education-num",  # being dropped
    "marital-status",  # clean
    "occupation",  # 1843 NaN out of 32560 samples
    "relationship",  # clean
    "race",  # clean
    "sex",  # clean
    "capital-gain",  # clean
    "capital-loss",  # clean
    "hours-per-week",  # clean
    "native-country",  # 583 NaN out of 32560 samples
    "income",  # clean
]


class CensusIncomeDataset(Dataset):
    def __init__(self, dataset, target_class, balance_dataset=True):
        """Construcs the Dataset given categorical US-Census-Income data.

        * Can also be used for sex classification task.

        Args:
            dataset (pandas.DataFrame): Census-Income Dataset.
            target_class (string): Target column name (e.g., "income" or "sex_target").

        Returns via __getitem__:
            label (torch.tensor): True target class
            x_input (torch.tensor): Census datapoint.
            mask (torch.tensor): Mask corresponding to datapoint.
            idx (int): Index corresponding to current items.
        """
        self.target_class = target_class
        self.labels = dataset[self.target_class].astype(np.int64).values
        self.dataset = dataset.drop(self.target_class, axis=1)

        if balance_dataset:
            under_sampler = RandomUnderSampler(random_state=42)
            self.dataset, self.labels = under_sampler.fit_resample(self.dataset, self.labels)

        self.size = self.dataset.shape[0]

        # Random initialization of masks | th.zeros performs faster on testset
        n_features = len(self.dataset.iloc[0].values)
        self.mask_merlin = th.zeros((self.size, n_features))
        self.mask_morgana = th.zeros((self.size, n_features))

    def update_mask(self, new_mask, mode, idx):
        """Replaces mask with optimized mask"""
        if mode == "merlin":
            self.mask_merlin[idx] = new_mask
        elif mode == "morgana":
            self.mask_morgana[idx] = new_mask
        else:
            raise ValueError(f"mode needs to be merlin or morgana, got: {mode}")

    def __len__(self):
        """Returns number of samples in dataset."""
        return self.size

    def __getitem__(self, idx):
        """Returns datapoints"""
        if self.target_class == "income":
            datapoint = self.dataset.iloc[idx].values
            self.x_input = datapoint[0]
            for row in datapoint[1:]:
                self.x_input = np.vstack((self.x_input, row))
            self.x_input = th.from_numpy(self.x_input).float()

        elif self.target_class == "sex_target":
            # target class is incorporated into datapoint
            datapoint = self.dataset.iloc[idx].values
            self.x_input = datapoint[0]
            for row in datapoint[1:]:
                self.x_input = np.vstack((self.x_input, row))
            self.x_input = th.from_numpy(self.x_input).float()

        return [self.labels[idx], self.x_input, self.mask_merlin[idx], self.mask_morgana[idx], idx]


def read_census_data(PATH, target_class, read_pre_processed=True, download=True):
    """Loads Dataset into Memory and removes useless columns.

    In addition, training and test datasets are adjusted to the required format.

    args:
    PATH (str): Path to data, which is not pre-processed.
    target_class (str): Target class name in Census DataFrame (e.g., "income" or "sex_target")
    read_pre_processed (bool, optional): If true, reads pre-processed data from local disk. Defaults to True.
    download (bool, optional): If true, the files are downloaded from the URL if they do not already exist.

    Returns:
        pd.DataFrame: Training and Test Dataset
    """

    if download == True:
        if os.path.isfile("data/adult.data") and os.path.isfile("data/adult.test"):
            print("Files already downloaded")
        else:
            sh.mkdir("data")
            url_train = "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data"
            url_test = "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test"
            print("Downloading Train and Test Dataset")
            wget.download(url_train, out="data")
            wget.download(url_test, out="data")

    if read_pre_processed:
        # Read preprocessed data
        train_data = pd.read_pickle(f"data/{target_class}_encoded_data_train.pkl")
        test_data = pd.read_pickle(f"data/{target_class}_encoded_data_test.pkl")
    else:
        # Read dataframe and apply preprocessing (takes some seconds)
        df_train_data = pd.read_csv(f"{PATH}.data", sep=",", header=None, names=header_names, na_values=' ?')
        df_train_data = df_train_data.dropna()
        df_train_data = df_train_data.drop(columns=["fnlwgt", "education"], axis=1)
        
        df_test_data = pd.read_csv(f"{PATH}.test", sep=",", header=None, names=header_names, na_values=' ?')
        df_test_data = df_test_data.dropna()
        df_test_data = df_test_data.drop(columns=["fnlwgt", "education"], axis=1)
        
        # Remove "." after each entry of income in test set
        df_test_data["income"] = (
            df_test_data["income"].astype(str).str.replace(r".", r"", regex=False)
        )
        if target_class == "sex_target":
            # Remove features that are strong correlated with sex-feature
            df_train_data = df_train_data.drop(columns=["relationship", "marital-status"], axis=1)
            df_test_data = df_test_data.drop(columns=["relationship", "marital-status"], axis=1)

        train_data, test_data = preprocess_data(df_train_data, df_test_data, target_class)

    return train_data, test_data


def preprocess_data(train_data, test_data, target_class, write_to_disk=True):
    """Preprocessing/Loading of the Datasets.

    * Applies customized one-hot-encodings. Fixed length of one-hot-encoded categorical features!

    * Each continuous feature is represented as a vector of same length like the one-hot-encodings,
      but each entry of the vector is set to the constant value of the continuous variable.

    * Leads to small performance drop wrt accuracy -- probably due to lengthening of the continuous variables,
      but can be compensated by more epochs.

    * Target class is either "income" or "sex".

    Args:
        train_data (pd.DataFrame): Train Dataset as DataFrame.
        test_data (pd.DataFrame): Test Dataset as DataFrame.
        target_class (str): Target class name in Census DataFrame (e.g., "income" or "sex_target").
        write_to_disk (bool, optional): If true, saves the pre-processed dataframe to local dir. Defaults to True.

    Returns:
        train_data (pd.DataFrame): Preprocessed One-Hot-Encoded Train Dataset
        test_data (pd.DataFrame): Preprocessed One-Hot-Encoded Test Dataset
    """
    print("... Preprocess Data and store locally ...")
    if target_class == "sex_target":
        # Add another column serving as target column
        train_data.insert(11, "sex_target", train_data["sex"], allow_duplicates=True)
        test_data.insert(11, "sex_target", test_data["sex"], allow_duplicates=True)

        # Binary encoding for sex
        label_encoder = LabelBinarizer(neg_label=-1, pos_label=1)
        train_data["sex"] = label_encoder.fit_transform(train_data["sex"])
        test_data["sex"] = label_encoder.transform(test_data["sex"])

    # Split columns into categorical_columns / categorical_train_columns / continuous_columns
    categorical_columns = [cat for cat in train_data.columns if type(train_data[cat][0]) == str]

    # Separate categorical train features from target class
    categorical_train_columns = [cat for cat in categorical_columns if cat != target_class]
    continuous_columns = [col for col in train_data.columns if col not in categorical_columns]
    
    scale_columns = continuous_columns[:]
    if target_class == "sex_target":    
        scale_columns.remove('sex')

    scaler = MinMaxScaler()
    train_data[scale_columns] = scaler.fit_transform(train_data[scale_columns])
    test_data[scale_columns] = scaler.transform(test_data[scale_columns])

    # Apply one-hot-encodings to categorical features and replace columns with encodings
    one_hot_encoders = {}
    for cat_col in categorical_train_columns:
        one_hot_encoders[cat_col] = OneHotEncoder(sparse=False)
        traincol_encodings = (
            one_hot_encoders[cat_col]
            .fit_transform(train_data[cat_col].values.reshape(-1, 1))
            .tolist()
        )
        train_data[cat_col] = traincol_encodings
        testcol_encodings = (
            one_hot_encoders[cat_col].transform(test_data[cat_col].values.reshape(-1, 1)).tolist()
        )
        test_data[cat_col] = testcol_encodings

    # Encode target classes
    label_encoder = LabelBinarizer()
    train_data[target_class] = label_encoder.fit_transform(train_data[target_class])
    test_data[target_class] = label_encoder.transform(test_data[target_class])

    # "native-country" is class with most categories
    max_encoding_size = len(one_hot_encoders["native-country"].categories_[0])

    train_data = convert_dataset_to_correct_format(
        train_data, continuous_columns, categorical_train_columns, max_encoding_size
    )
    test_data = convert_dataset_to_correct_format(
        test_data, continuous_columns, categorical_train_columns, max_encoding_size
    )

    if write_to_disk:
        train_data.to_pickle(path=f"data/{target_class}_encoded_data_train.pkl")
        test_data.to_pickle(path=f"data/{target_class}_encoded_data_test.pkl")

    return train_data, test_data


def convert_dataset_to_correct_format(
    dataset, continuous_columns, categorical_train_columns, max_encoding_size
):
    """Converts UCI Census Dataset to the required format for ArthurMerlin framework.

    Resizes continuous variables to vector of fixed length, where each entry is the continuous variable.
    Pads one-hot-encoded categorical data with zeros to same fixed length. This leads to
    repetition of information, making it more difficult for the model to filter out
    patterns from the data.

    Args:
        dataset (pd.DataFrame): UCI Census Data
        continuous_columns (list): List of str containing the headers of continuous features.
        categorical_train_columns (list): List of str containing the headers of categorical features.
        max_encoding_size (int): Max length of one-hot-encoded features along all features.

    Returns;
        dataset (pd.DataFrame): Converted dataset.
    """

    for category in continuous_columns:
        dataset[category] = dataset.apply(
            lambda row: np.resize(row[category], new_shape=max_encoding_size), axis=1
        )
    for category in categorical_train_columns:
        dataset[category] = dataset.apply(lambda row: np.array(row[category]), axis=1)

    for category in categorical_train_columns:
        dataset[category] = dataset.apply(
            lambda row: np.pad(
                row[category],
                pad_width=(0, max_encoding_size - row[category].shape[0]),
                constant_values=0,
            ),
            axis=1,
        )
    return dataset
