import torch
from torch.utils.data import DataLoader

import sh
from arthur_merlin_framework import adversarial_training
from census_data import CensusIncomeDataset, read_census_data
from model_training import ArthurClassifier
from utilities import load_pretrained, seed_worker

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print("Running on Device:", device)

g = torch.Generator()
g.manual_seed(40)

sh.rm("-r", "-f", "runs")
sh.mkdir("runs")

def main():
    # Download and preprocess US Census Income dataset
    train_data, test_data = read_census_data(
        PATH="data/adult", target_class="sex_target", read_pre_processed=False, download=True
    )

    train_data = CensusIncomeDataset(dataset=train_data, target_class="sex_target")
    test_data = CensusIncomeDataset(dataset=test_data, target_class="sex_target")

    # Train Dataloader
    train_dataloader = DataLoader(
        train_data,
        batch_size=512,
        shuffle=True,
        num_workers=4,
        worker_init_fn=seed_worker,
        generator=g,
    )
    # Test Dataloader
    test_dataloader = DataLoader(
        test_data,
        batch_size=512,
        shuffle=False,
        num_workers=4,
        worker_init_fn=seed_worker,
        generator=g,
    )

    n_features, encoding_size = train_data[0][1].shape
    # Initialize Neural Network
    model = ArthurClassifier(input_size=n_features * encoding_size, output_size=3).to(device)

    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("Number of Parameters: ", n_params)

    # (gamma, alpha)
    params = ((0, 3.5), (2, 3.5), (2, 10), (0, 0))

    for gamma, alpha in params:
        # Load pretrained model
        PATH = "pretrained_model/best_checkpoint.pth"
        model = load_pretrained(model, PATH)
        print(f"Start training with gamma={gamma} and alpha={alpha}")
        adversarial_training(
            model,
            train_data,
            test_dataloader,
            device,
            mask_size=1,
            nepochs=20,
            lr=5e-1,
            gamma=gamma,
            alpha=alpha,
        )


if __name__ == "__main__":
    main()
