import math

import torch
import itertools
import numpy
from custom_criterion import custom_criterion
from metrics import get_accuracy, get_bce_distortion, get_soundness
from utilities import EarlyStopping, apply_mask


class SFW(torch.optim.Optimizer):
    """Stochastic Frank Wolfe Algorithm

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining parameter groups
        learning_rate (float): learning rate between 0.0 and 1.0
        rescale (string or None): Type of learning_rate rescaling. Must be 'diameter', 'gradient' or None
        momentum (float): momentum factor, 0 for no momentum
    """

    def __init__(self, params, learning_rate=0.1, rescale="diameter", momentum=0.9):
        if not (0.0 <= learning_rate <= 1.0):
            raise ValueError("Invalid learning rate: {}".format(learning_rate))
        if not (0.0 <= momentum <= 1.0):
            raise ValueError("Momentum must be between [0, 1].")
        if not (rescale in ["diameter", "gradient", None]):
            raise ValueError("Rescale type must be either 'diameter', 'gradient' or None.")

        # Parameters
        self.rescale = rescale

        defaults = dict(lr=learning_rate, momentum=momentum)
        super(SFW, self).__init__(params, defaults)

    @torch.no_grad()
    def step(self, constraints, closure=None):
        """Performs a single optimization step.
        Args:
            constraints (iterable): list of constraints, where each is an initialization of Constraint subclasses
            parameter groups
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                d_p = p.grad

                # Add momentum
                momentum = group["momentum"]
                if momentum > 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        param_state["momentum_buffer"] = d_p.detach().clone()
                    else:
                        param_state["momentum_buffer"].mul_(momentum).add_(d_p, alpha=1 - momentum)
                        d_p = param_state["momentum_buffer"]

                v = constraints[idx].lmo(d_p)  # LMO optimal solution

                if self.rescale == "diameter":
                    # Rescale lr by diameter
                    factor = 1.0 / constraints[idx].get_diameter()
                elif self.rescale == "gradient":
                    # Rescale lr by gradient
                    factor = torch.norm(d_p, p=2) / torch.norm(p - v, p=2)
                else:
                    # No rescaling
                    factor = 1

                lr = max(0.0, min(factor * group["lr"], 1.0))  # Clamp between [0, 1]

                p.mul_(1 - lr)
                p.add_(v, alpha=lr)
                idx += 1
        return loss


class PositiveKSparsePolytope:
    """Polytopes with vertices v \in {0, 1}^n such that exactly k entries are nonzero.

    This is exactly the intersection of B_1(k) with B_inf(1) and {x | x>=0}
    Note: Here the dimension is n, but the dimension of the oracle input/solution is bs*n
    """

    def __init__(self, n, bs, k=1):
        self.k = min(k, n)
        self.diameter = math.sqrt(2 * k) if 2 * k <= n else math.sqrt(n)

    @torch.no_grad()
    def get_diameter(self):
        return self.diameter

    @torch.no_grad()
    def lmo(self, x):
        """Returns v in PositiveKSparsePolytope minimizing v*x"""
        # NOTE: we have to do this per image
        v = torch.zeros_like(x)
        minIndices = torch.topk(x.flatten(start_dim=1), k=self.k, largest=False).indices
        v.flatten(start_dim=1).scatter_(1, minIndices, 1.0)
        v[x >= 0] = 0.0

        return v

    @torch.no_grad()
    def shift_inside(self, x, check_necessity=False):
        """
        Projects x to the PositiveKSparsePolytope.
        This is a valid projection, although not the one mapping to minimum distance points.

        Args:
            x (torch.tensor): Input data.
            check_necessity (bool, optional): _description_. Defaults to False.

        Returns:
            v: Projected input onto k-sparse vertices of polytope.
        """
        # Check if necessary
        if check_necessity:
            l1 = torch.norm(x.flatten(start_dim=1), p=1, dim=1)
            linf = torch.norm(x, p=float("inf"))
            if linf <= 1 and (l1 <= self.k).all() and (x >= 0).all():
                # Is in the polytope, no need to shift
                return x

        # For now, we just take the closest vertex
        # Note: we have to do this per image
        v = torch.zeros_like(x).flatten(start_dim=1)
        maxIndices = torch.topk(torch.abs(x.flatten(start_dim=1)), k=self.k).indices.to(x.device)
        v.scatter_(1, maxIndices, 1.0)
        v.requires_grad = True

        return v.reshape(x.shape)


def opt_mask(
    model,
    x_input,
    y_target,
    mode,
    k,
    mask,
    device,
    max_iter=1500,
    lr=1e-1,
    momentum=0.9,
    stoptol=1e-5,
    verbose=False,
    alpha=0,
):
    """Optimizes the mask and minimizes the distortion via Stochastic-Frank-Wolfe (SFW).

    Calculates the distortion between the logits of the real and (non-projected) masked input datapoints.
    Reduces the distortions by optimizing the criterion/loss. Regularizer with penalty term incorporated.
    Represents either Merlin or Morgana. Early stopping with two criteria incorporated.
    Returns masked input, k-sparse mask and metrics.

    Args:
        model (torch.nn.model): Classifier.
        x (torch.tensor): Input datapoints (batch).
        y (torch.tensor): True class of input (batch).
        mode (str): "morgana" or "merlin".
        k (int): Size of mask.
        mask (torch.tensor): k-sparse mask of datapoint (batch).
        device (str): Cuda or CPU.
        max_iter (int, optional): Maximal opimization steps before termination. Defaults to 1500.
        lr (float, optional): Learning rate of SFW optimizer. Defaults to 0.1.
        momentum (float, optional): Momentum of SFW optimizer. Defaults to 0.9.
        stoptol (float, optional): Early stopping criteria. Defaults to 1e-5.
        verbose (bool, optional): Print more information per optimization step. Defaults to False.
        penalty (float, optional): Penalty for regularization. Defaults to 0.25.
        alpha (float, optional): Factor for penalty.

    Returns:
        x_masked (torch.tensor): (k-spare) masked datapoint (batch).
        mask (torch.tensor): (k-sparse) optimized masks (batch).
        metrics (dict): Dictionary of all relevant metrics on (k-sparse) masked input.
    """
    batch_size = y_target.shape[0]
    assert x_input.numel() % batch_size == 0  # check if remainder is zero
    output_size = model.output_size

    if output_size == 1:
        # Binary classification problem (only merlin is incorporated in framework)
        mask_criterion = torch.nn.BCEWithLogitsLoss
    elif output_size == 3:
        # Non-binary classification problem (both morgana or merlin are incorporated in framework)
        if mode == "merlin":
            mask_criterion = torch.nn.CrossEntropyLoss()
        elif mode == "morgana":
            mask_criterion = custom_criterion

    # Set constraint for optimisation
    constraint = PositiveKSparsePolytope(n=x_input.numel() / batch_size, bs=batch_size, k=k)
    # NOT ACTIVATED: Will only do something if necessary
    mask = constraint.shift_inside(mask, check_necessity=False)
    # Do not push nn.Parameter() to cuda, since this makes s non-leaf
    mask = torch.nn.Parameter(mask.to(device), requires_grad=True)
    mask.requires_grad = True

    rde_optim = SFW([mask], learning_rate=lr, momentum=momentum, rescale=None)

    # Disable the network's autograd
    for param in model.parameters():
        param.requires_grad = False

    # Early stopping variable
    stopping_criteria = EarlyStopping(tolerance=1, min_delta=stoptol)
    distortion_old = float("inf")
    metrics = dict()

    iter = 0
    while not stopping_criteria.early_stop:
        # Optimization of mask
        rde_optim.zero_grad()
        distortion = get_bce_distortion(model, x_input, y_target, mask, criterion=mask_criterion)
        gender_regularizer = regularize_gender_feature(mask, alpha)
        mask_loss = distortion + gender_regularizer
        mask_loss.backward()
        # mask is not k-sparse after rde_optim.step(constraints=[constraint])
        # torch.nn.utils.clip_grad_norm_(mask, 1)
        rde_optim.step(constraints=[constraint])

        if verbose and (iter % 100 == 0 or iter == 0):
            # Give verbose updates of accuracy
            with torch.no_grad():
                l_0 = (mask > 1e-30).sum().item()
                sparsity = 1 - l_0 / float(mask.numel())
                l_1 = torch.norm(mask, p=1).item()

                # unconstrained accuracy
                logits_unconstrained = model.eval()(apply_mask(x_input, mask))
                acc_unconstrained = get_accuracy(logits=logits_unconstrained, y_target=y_target)
                # constrained accuracy
                mask_constrained = constraint.shift_inside(mask, check_necessity=False)
                logits_constrained = model.eval()(apply_mask(x_input, mask_constrained))
                acc_constrained = get_accuracy(logits=logits_constrained, y_target=y_target)

                print(
                    f"Iteration: {iter}: | "
                    f"Distortion: {distortion.item():.3f} | "
                    f"Unconstrained Acc.: {acc_unconstrained:.4f} | "
                    f"Constrained Acc.: {acc_constrained:.4f} | "
                    f"Avg. Sparsity: {sparsity:.2f} | "
                    f"L_1: {l_1:.3f}"
                )

        # Check stopping criteria
        iter_crit = (iter + 1) > max_iter

        if iter_crit or stopping_criteria(distortion.item(), distortion_old):
            stopping_criteria.early_stop = True
            if verbose:
                print(f"Stopping mask optimization after iteration: {iter}")

        distortion_old = distortion.item()
        iter += 1

    with torch.no_grad():
        # project onto vertices of k-sparse polytope
        mask = constraint.shift_inside(mask, check_necessity=False)
        x_masked = apply_mask(x_input, mask)
        logits = model.eval()(x_masked)

        # Get accuracy of k-sparse mask
        if mode == "merlin":
            acc = get_accuracy(logits, y_target, output_size=output_size)
            metrics["accuracy"] = acc
        elif mode == "morgana":
            soundness = get_soundness(logits, y_target)
            metrics["soundness"] = soundness
        else:
            raise ValueError(f"mode needs to be either merlin or morgana, got {mode}")
        l_0 = (mask > 1e-30).sum().item()
        sparsity = 1 - l_0 / float(mask.numel())

        metrics["it_until_conv"] = iter - 1
        metrics["distortion"] = distortion.item()
        metrics["sparsity"] = sparsity

    # Enable the network's autograd
    for param in model.parameters():
        param.requires_grad = True

    return (
        (x_masked.detach().clone()),
        mask.detach().clone(),
        metrics,
    )


def regularize_gender_feature(mask, alpha=0):
    """Introduces regularization to avoid sex as a feature.

    Args:
        mask (torch.tensor): (non-sparse) mask.
        penalize (bool, optional): If true, activates regularization.
        alpha (int, optional): Multiplier. Defaults to 0.

    Returns:
        gender_regularizer (float): Loss regularizer on gender feature.
    """
    # Regularizes mask optimization to avoid gender feature
    gender_regularizer = alpha * torch.mean(mask[:, 5])
    return gender_regularizer


def brute_force_mask(
    model,
    x_input,
    y_target,
    mode,
    k,
    device,
):
    """Finds masks for each datapoint with lowest morgana loss.

    Args:
        model (torch.nn.model): Classifier.
        x_input (torch.tensor): Input datapoints (batch).
        y_target (torch.tensor): True class of input (batch).
        mode (str): "morgana" or "merlin".
        k (int): Size of mask.
        device (str): Cuda or CPU.

    Returns:
        x_masked (torch.tensor): (k-spare) masked datapoint (batch).
        mask (torch.tensor): (k-sparse) masks found with brute-force.
    """
    batch_size = x_input.shape[0]
    n_features = x_input.shape[1]

    if mode == "merlin":
        mask_criterion = torch.nn.CrossEntropyLoss(reduction="none")
    elif mode == "morgana":
        mask_criterion = custom_criterion

    with torch.no_grad():
        # Create all possible combinations of k-sparse masks for brute force search
        which = numpy.array(list(itertools.combinations(range(n_features), k)))
        grid = numpy.zeros((len(which), n_features), dtype="float32")
        grid[numpy.arange(len(which))[None].T, which] = 1
        grid = torch.tensor(grid)

        # Initialize first k-sparse mask
        mask = grid[0]
        mask = mask.repeat(batch_size, 1).to(device)
        
        x_masked = apply_mask(x_input, mask)
        logits = model.eval()(x_masked)
        loss = mask_criterion(logits, y_target, reduction="none")

        for new_mask in grid[1:]:
            # Iterate all combinations with brute force
            new_mask = new_mask.repeat(batch_size, 1).to(device)
            x_masked = apply_mask(x_input, new_mask)
            logits = model.eval()(x_masked)
            new_loss = mask_criterion(logits, y_target, reduction="none")
            loss_diff = loss - new_loss
            updated_loss = torch.where(loss_diff > 0, new_loss, loss)
            # Update mask entries, where necessary
            mask[~updated_loss.eq(loss)] = new_mask[~updated_loss.eq(loss)]
            # Update loss, where necessary
            loss = updated_loss
        # final masking
        x_masked = apply_mask(x_input, mask)
    return x_masked, mask
