import torch as th

from utilities import apply_mask


def get_bce_distortion(model, x, y_target, mask, criterion):
    """Calulates the loss of the masked input fed to Arthur wrt the real class.

    Masking is done either by Merlin or Morgana. The corresponding loss-criterion is passed to calculate
    the correct loss.

    Args:
        model (torch.nn.model): Classifier.
        x (torch.tensor): Input features.
        y_target (torch.tensor): Target classes.
        mask (torch.tensor): Mask either by Merlin or Morgana.
        criterion (torch.nn): Loss-Criterion corresponding to Merlin or Morgana.

    Returns:
        distortion (float): Distortion between logits of masked input and target class.
    """
    x_masked = apply_mask(x, mask)
    phi_x_masked = model.eval()(x_masked)
    distortion = criterion(phi_x_masked, y_target)
    return distortion


@th.no_grad()
def get_accuracy(logits, y_target, output_size):
    """Calulates the accuracy of the model's output (either masked or real).

    Args:
        x (torch.tensor): Input datapoints (batch)
        y (torch.tensor): True class of input (batch)
        output_size (int): Output size of Arthur (dependent on number of classes)

    Returns:
        accuracy (float): Average accuracy of batch
    """
    if output_size == 1:
        preds = th.sigmoid(logits)
        preds = preds.round()
    elif output_size == 3:
        preds = th.softmax(logits, dim=1)
        preds = preds.argmax(dim=1)
    return preds.eq(y_target.view_as(preds)).sum().item() / float(len(y_target))


@th.no_grad()
def get_soundness(logits, y_target):
    """Returns soundness wrt Morgana's masked inputs.

    Args:
        logits (torch.Tensor): Arthur's logits corresponding to Morgana's masked input.
        y_target (torch.Tensor): True target labels.

    Returns:
        soundness (torch.Tensor): Accuracy wrt to true target class or "I don't know"
    """
    preds = th.softmax(logits, dim=1)
    preds = preds.argmax(dim=1)
    correct_class = preds.eq(y_target.view_as(preds)).sum().item()
    idk = preds.eq(2).sum().item()
    soundness = (correct_class + idk) / float(len(y_target))
    return soundness


def print_metrics(prefix, metrics):
    """Printing results of coop_training function.

    Args:
        prefix (str): Prefix, e.g., "train" or "test".
        metrics (dict): Dictionary containing results.
    """
    for key in metrics.keys():
        message = prefix +  key + f": {metrics[key]:1.3f}"
        print(message)
