import torch as th
from torch.utils.tensorboard import SummaryWriter
from torch import nn
from tqdm import tqdm

from utilities import get_lr, print_epoch_results


class ArthurClassifier(nn.Module):
    def __init__(self, input_size, output_size=3):
        """Simple nn to classify categorical data.

        Args:
            input_size (int): Number of total input dimensions
            output_size (int): Number of Target Classes
        """
        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.fw_layers = nn.Sequential(
            nn.Linear(input_size, 50),
            nn.ReLU(),
            nn.Linear(50, self.output_size),
        )

    def forward(self, x_input):
        x_input = x_input.flatten(start_dim=1)
        x = self.fw_layers(x_input)
        return x


def regular_training(
    model, train_data, train_dataloader, test_data, test_dataloader, n_epochs, device
):
    """Regular supervised training routine for ArthurClassifier.

    Args:
        model (torch.nn.Module): Classifier.
        train_data (torch.data): Training Set.
        train_dataloader (torch.dataloader): Iterator.
        test_data (torch.data): Test Set.
        test_dataloader (torch.dataloader): Iterator.
        n_epochs (int): Number of epochs.
        device (str): CPU or Cuda.
    """
    # If true, train ArthurClassifier without Merlin and Morgana for multiple epochs
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    optimizer = th.optim.Adam(model.parameters(), lr=1e-4)
    log_dir = (
        f"runs/Census-Model-{n_params:1.0E}-Params-{n_epochs}Epochs-lr_{get_lr(optimizer):1.1E}"
    )
    logger = SummaryWriter(log_dir)
    best_score = 0
    criterion = th.nn.CrossEntropyLoss()
    for epoch in range(n_epochs):
        train_loss, train_acc = train_epoch(
            model, optimizer, criterion, train_dataloader, train_data, logger, epoch, device
        )
        test_loss, test_acc, best_score = evaluate_epoch(
            model,
            optimizer,
            criterion,
            test_dataloader,
            test_data,
            logger,
            epoch,
            best_score,
            device,
        )
        print_epoch_results(epoch, train_loss, train_acc, test_loss, test_acc)

    # Seperately log hyperparameters | Useful for hyperparameter tuning
    log_hyperparameters(
        logger,
        n_epochs,
        n_params,
        best_score,
        learning_rate=get_lr(optimizer),
        batch_size=train_dataloader.batch_size,
    )
    logger.close()


def train_epoch(model, optimizer, criterion, train_dataloader, train_data, logger, epoch, device):
    """Training routine w/o masks.

    Calculates average loss and accuracy of an entire epoch. Logs everything on tensorboard.

    Args:
        model (nn.Module): Classifier.
        optimizer (torch.optim): Optimizer (e.g., SGD or Adam).
        criterion (torch.nn.BCEWithLogitsLoss): Binary-cross-entropy loss with logits.
        dataloader (torch.utils.data.Dataloader): Train dataloader.
        logger (tensorboard): Logger for metrics (e.g., accuracy and loss).
        epoch (int): Number of epochs.
        device (str): Cuda or CPU.

    Returns:
        train_loss (float): Average loss.
        train_acc (float): Average accuracy.
    """
    model.train()
    correct = 0
    epoch_loss = 0
    for y, x_input, _, _, _ in tqdm(train_dataloader):
        x_input = x_input.to(device)
        y = y.to(device)
        # Forward Pass
        output = model(x_input)
        loss = criterion(output, y)
        epoch_loss += loss
        # accuracy
        preds = th.softmax(output, dim=1)
        correct += (preds.argmax(dim=1) == y).type(th.float).sum().item()
        # Backward pass and optimization of model parameters
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    train_acc = correct / len(train_data)
    train_loss = epoch_loss / len(train_dataloader)
    logger.add_scalar("Loss/train", train_loss, epoch)
    logger.add_scalar("Accuracy/train", train_acc, epoch)

    return train_loss, train_acc


def evaluate_epoch(
    model, optimizer, criterion, test_dataloader, test_data, logger, epoch, best_score, device
):
    """Evaluation Routine for model w/o mask.

    Evaluates model on entire epoch and calculates average accuracy and loss.
    Logs everything on tensorboard. No mask input.

    Args:
        model (nn.Module): Classifier.
        optimizer (torch.optim): Optimizer (e.g., SGD or Adam).
        criterion (torch.nn.BCEWithLogitsLoss): Binary-cross-entropy loss with logits.
        dataloader (torch.utils.data.Dataloader): Test dataloader.
        logger (tensorboard): Logger for metrics (e.g., accuracy and loss).
        epoch (int): Number of epochs.
        best_score (float): Best validation accuracy achieved so far.
        device (str): Cuda or CPU.

        Returns:
            test_loss (float): Average loss.
            test_acc (float): Average accuracy.
            best_score (float): Updated best validation accuracy achieved so far.
    """
    model.eval()
    correct = 0
    epoch_loss = 0
    with th.no_grad():
        for y, x_input, _, _, _ in tqdm(test_dataloader):
            x_input = x_input.to(device)
            y = y.to(device)
            # Forward Pass
            output = model(x_input)
            loss = criterion(output, y)
            epoch_loss += loss
            # accuracy
            preds = th.softmax(output, dim=1)
            correct += (preds.argmax(dim=1) == y).type(th.float).sum().item()
        test_acc = correct / len(test_data)
        test_loss = epoch_loss / len(test_dataloader)
        logger.add_scalar("Loss/test", test_loss, epoch)
        logger.add_scalar("Accuracy/test", test_acc, epoch)
        logger.flush()

        if test_acc >= best_score:
            # Save model with highest accuracy
            th.save(
                dict(
                    model=model.state_dict(),
                    optimizer=optimizer.state_dict(),
                    epoch=epoch,
                    best_score=test_acc,
                    batch_size=test_dataloader.batch_size,
                    learning_rate=get_lr(optimizer),
                ),
                f"{logger.log_dir}/best_checkpoint.pth",
            )
            best_score = test_acc

    return test_loss, test_acc, best_score


def log_hyperparameters(logger, n_epochs, n_params, best_score, learning_rate, batch_size):
    """Logs hyperparameters after training and validation is finished for all epochs.

    Args:
        logger (tensorboard): Logger for hyperparameters (e.g., batch_size, epochs, etc.)
        batch_size (int): Batch size of dataloader.
        n_epochs (int): Number of epochs.
        n_params (int): Number of model parameters.
        learning_rate (float): Constant learning rate used for training.
        best_score (float): Best achieved validation accuracy of all epochs.
    """
    logger.add_hparams(
        hparam_dict={
            "hparam/lr": learning_rate,
            "hparam/batch_size": batch_size,
            "hparam/epochs": n_epochs,
            "hparam/model_parameters": n_params,
        },
        metric_dict={"hparam/test_acc": best_score},
    )
