import random

import matplotlib.pyplot as plt
import numpy as np
import torch

class Averager(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.sum = 0
        self.count = 0
        self.avg = 0

    def result(self):
        if self.count > 0:
            return self.avg
        else:
            return np.nan

    def add(self, val, n=1):
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class EarlyStopping:
    """Sets Early Stopping Criteria.

    If loss/rde increases or stagnates several times, early stop is triggered.

    Args:
        tolerance (int): Number of repeated times until early stopping is triggered.
        min_delta (float): Threshold value for increasing the counter.
    """

    def __init__(self, tolerance=3, min_delta=1e-5):
        self.tolerance = tolerance
        self.min_delta = min_delta
        self.counter = 0
        self.early_stop = False

    def __call__(self, val1, val2):
        if val1 - val2 > self.min_delta:
            # If loss increases several times, call early stop
            self.counter += 1
            if self.counter >= self.tolerance:
                self.early_stop = True
        elif abs(val1 - val2) < self.min_delta:
            # If the loss stagnates several times, an early stop is required.
            self.counter += 1
            if self.counter >= self.tolerance:
                self.early_stop = True
        else:
            self.counter = 0
        return self.early_stop


def apply_mask(x, mask):
    """Applies mask on inputs from census dataset via broadcasting.

    Neglecting the batch-dimension, a single datapoint x_input has the shape (12,42),
    while the mask is given as a vector of length (12,). Another dimension is added to the mask,
    i.e., it becomes (12,1), which enables broadcasting.

    Args:
        x (torch.tensor): Input datapoint.
        mask (torch.tensor): 12-dimensional mask.
    Returns:
        x_masked (torch.tensor): Masked input.

    """
    mask = mask.view(x.shape[0], x.shape[1], -1)
    return mask * x  # broadcasting


def print_epoch_results(epoch, train_loss, train_acc, test_loss, test_acc):
    """Prints results/metrics obtained within one entire epoch.

    Args:
        epoch (int): Number of epochs.
        train_loss (float): Logging metric.
        train_acc (float): Logging metric.
        test_loss (float): Logging metric.
        test_acc (float): Logging metric.
    """
    # Printing statements after each epoch
    print(
        f"Epoch: {epoch+1:02} | "
        f"Train Loss: {train_loss:.3f} | "
        f"Train Accuracy: {train_acc:.3f} | "
        f"Test Loss: {test_loss:.3f} | "
        f"Test Accuracy: {test_acc:.3f}"
    )


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group["lr"]


def load_pretrained(model, PATH):
    checkpoint = torch.load(PATH)
    model.load_state_dict(checkpoint["model"])

    return model


def plot_to_tensorboard(
    logger,
    merlin_feat_frac,
    morgana_feat_frac,
    metrics,
    feature_list,
    gamma,
    alpha,
    epoch,
    tag="Feature Selection Bar Plot",
):

    params = {
        "text.usetex": True,
        "font.size": 18,
        "font.family": "serif",
        "font.serif": "Times New Roman",
    }
    plt.rcParams.update(params)

    width = 0.4
    fig, ax = plt.subplots()

    textstr = "\n".join(
        (
            r"$Completeness=%.0f\%%$" % (100 * metrics["completeness"],),
            r"$Soundness=%.2f\%%$" % (100 * metrics["soundness"],) if gamma > 0 else "",
            r"$\gamma=%.1f$, \quad $\alpha=%.1f$" % (gamma, alpha),
        )
    )
    # place a text in upper left in axes coords
    ax.text(0.03, 0.98, textstr, transform=ax.transAxes, verticalalignment="top")
    xt = np.arange(len(feature_list))
    ax.set_ylim(bottom=0, top=1)

    ax.bar(
        xt - width / 2,
        merlin_feat_frac.flatten(),
        width,
        label="Chosen by Merlin",
        color="mediumturquoise",  # "mediumturquoise" for less saturation | "blue" for more saturation
    )

    if gamma > 0:
        ax.bar(
            xt + width / 2,
            morgana_feat_frac.flatten(),
            width,
            label="Chosen by Morgana",
            color="salmon",  # "salmon" for less saturation | "red" for more saturation
        )

    ax.set_xticks(xt, feature_list)
    ax.tick_params(axis="x", rotation=45)
    plt.setp(ax.xaxis.get_majorticklabels(), ha="right", rotation_mode="anchor")
    # ax.set_xlabel("Feature")
    ax.set_ylabel("Relative Occurence")
    # ax.set_title("Feature selection by Merlin and Morgana")
    # ax.legend(loc="best")
    ax.grid(axis="y", alpha=0.3)
    fig.tight_layout()
    # plt.savefig("merlin_feature_fractions.pdf", transparent=True, edgecolor="none")
    logger.add_figure(tag=tag, figure=fig, global_step=epoch)
    logger.flush()


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)
